"""Sandbox, buffer, and volume tool definitions.

These tools run code inside the Modal sandbox or manage buffer/volume
operations. They are built by :func:`build_sandbox_tools` and merged
into the main tool list by :func:`~fleet_rlm.react.tools.build_tool_list`.

RLM delegation tools have been extracted to tools_rlm_delegate.py and
memory intelligence tools to tools_memory_intelligence.py as part of
the modularization effort (Linear: QRE-273).
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
import logging

from . import execute_submit
from .memory_intelligence import build_memory_intelligence_tools
from .delegate import build_rlm_delegate_tools
from .sandbox_helpers import _resolve_volume_path

if TYPE_CHECKING:
    from ..agent import RLMReActChatAgent


@dataclass(slots=True)
class _SandboxToolContext:
    """Shared context for sandbox/volume tool operations."""

    agent: "RLMReActChatAgent"


def _execute_submit_ctx(
    ctx: _SandboxToolContext, code: str, *, variables: dict[str, Any] | None = None
) -> dict[str, Any]:
    return execute_submit(ctx.agent, code, variables=variables or {})


def _resolve_path_or_error(
    *,
    path: str,
    default_root: str,
    allowed_root: str = "/data",
) -> tuple[str | None, dict[str, Any] | None]:
    try:
        return (
            _resolve_volume_path(
                path,
                default_root=default_root,
                allowed_root=allowed_root,
            ),
            None,
        )
    except ValueError as exc:
        return None, {"status": "error", "error": str(exc)}


def _reload_volume_best_effort(ctx: _SandboxToolContext) -> None:
    if ctx.agent.interpreter._volume:
        try:
            ctx.agent.interpreter.reload()
        except Exception as exc:
            logging.exception("Best-effort volume reload failed: %s", exc)


def _commit_volume_best_effort(ctx: _SandboxToolContext) -> None:
    if ctx.agent.interpreter._volume:
        try:
            ctx.agent.interpreter.commit()
        except Exception as exc:
            logging.exception("Best-effort volume commit failed: %s", exc)


# ---------------------------------------------------------------------------
# Tool factory
# ---------------------------------------------------------------------------


def build_sandbox_tools(
    agent: "RLMReActChatAgent",
) -> list[Any]:
    """Build sandbox / buffer / volume tools bound to *agent*.

    Returns a list of ``dspy.Tool`` wrappers ready to be appended to the
    main tool list built by ``build_tool_list``.
    """
    ctx = _SandboxToolContext(agent=agent)
    tools: list[Any] = []

    # -- RLM delegation tools (extracted to tools_rlm_delegate.py) ------------
    tools.extend(build_rlm_delegate_tools(agent))

    # -- Memory intelligence tools (extracted to tools_memory_intelligence.py) -
    tools.extend(build_memory_intelligence_tools(agent))

    # -- Sandbox editing -----------------------------------------------------

    def edit_file(path: str, old_snippet: str, new_snippet: str) -> dict[str, Any]:
        """Robustly edit a file by finding and replacing a unique text snippet.

        Fails if the old_snippet is not found or is not unique in the file.
        Use this over fragile `sed` commands for precise code editing.
        """
        code = """
try:
    with open(path, "r", encoding="utf-8") as f:
        content = f.read()
except FileNotFoundError:
    SUBMIT(status="error", error=f"File not found: {path}")
    exit(0)

count = content.count(old_snippet)
if count == 0:
    SUBMIT(status="error", error="old_snippet not found in file")
elif count > 1:
    SUBMIT(status="error", error=f"old_snippet is ambiguous (found {count} times)")
else:
    new_content = content.replace(old_snippet, new_snippet)
    with open(path, "w", encoding="utf-8") as f:
        f.write(new_content)
    SUBMIT(status="ok", path=path, message="File updated successfully")
"""
        return _execute_submit_ctx(
            ctx,
            code,
            variables={
                "path": path,
                "old_snippet": old_snippet,
                "new_snippet": new_snippet,
            },
        )

    # -- Buffer & volume management ------------------------------------------

    def read_buffer(name: str) -> dict[str, Any]:
        """Read the full contents of a sandbox buffer."""
        result = _execute_submit_ctx(
            ctx,
            "SUBMIT(items=get_buffer(name))",
            variables={"name": name},
        )
        items = result.get("items", [])
        return {"status": "ok", "name": name, "items": items, "count": len(items)}

    def clear_buffer(name: str = "") -> dict[str, Any]:
        """Clear one sandbox buffer (or all buffers when name is empty)."""
        if name:
            code = 'clear_buffer(name)\nSUBMIT(status="ok", scope="single", name=name)'
            variables: dict[str, Any] = {"name": name}
        else:
            code = 'clear_buffer()\nSUBMIT(status="ok", scope="all")'
            variables = {}
        return _execute_submit_ctx(ctx, code, variables=variables)

    def save_buffer_to_volume(name: str, path: str) -> dict[str, Any]:
        """Persist a sandbox buffer to Modal Volume storage as JSON."""
        resolved_path, error = _resolve_path_or_error(
            path=path,
            default_root="/data/workspace/buffers",
        )
        if error is not None:
            return error

        code = """
import json
items = get_buffer(name)
payload = json.dumps(items, indent=2, ensure_ascii=False, default=str)
saved_path = save_to_volume(path, payload)
SUBMIT(status="ok", saved_path=saved_path, item_count=len(items))
"""
        result = _execute_submit_ctx(
            ctx,
            code,
            variables={"name": name, "path": resolved_path},
        )
        if result.get("status") == "ok":
            _commit_volume_best_effort(ctx)
        return result

    def load_text_from_volume(path: str, alias: str = "active") -> dict[str, Any]:
        """Load text from Modal Volume into host-side document memory."""
        resolved_path, error = _resolve_path_or_error(
            path=path,
            default_root="/data/workspace",
        )
        if error is not None:
            return error

        # Pull latest writes from other containers before reading.
        _reload_volume_best_effort(ctx)

        result = _execute_submit_ctx(
            ctx,
            'text = load_from_volume(path)\nSUBMIT(status="ok", text=text)',
            variables={"path": resolved_path},
        )
        text = str(result.get("text", ""))
        if text.startswith("[error:"):
            return {"status": "error", "error": text, "path": resolved_path}
        ctx.agent._set_document(alias, text)
        ctx.agent.active_alias = alias
        return {
            "status": "ok",
            "alias": alias,
            "path": resolved_path,
            "chars": len(text),
            "lines": len(text.splitlines()),
        }

    def process_document(path: str, alias: str = "active") -> dict[str, Any]:
        """Load a document from volume and register it for downstream analysis."""
        loaded = load_text_from_volume(path, alias=alias)
        if loaded.get("status") != "ok":
            return loaded
        text = ctx.agent.documents.get(alias, "")
        return {
            "status": "ok",
            "alias": alias,
            "path": loaded.get("path", path),
            "chars": len(text),
            "lines": len(text.splitlines()),
            "hint": "Use analyze_long_document or summarize_long_document for semantic processing.",
        }

    # -- Persistent memory management ----------------------------------------
    # These tools allow the agent to use the mounted volume as a persistent
    # "hard drive" for storing user profiles, archival documents, etc.

    def memory_read(path: str) -> dict[str, Any]:
        """Read a file from persistent memory (Modal Volume)."""
        resolved_path, error = _resolve_path_or_error(
            path=path,
            default_root="/data/memory",
        )
        if error is not None:
            return error

        _reload_volume_best_effort(ctx)

        code = """
try:
    with open(path, "r", encoding="utf-8") as f:
        content = f.read()
    SUBMIT(status="ok", path=path, content=content, chars=len(content))
except FileNotFoundError:
    SUBMIT(status="error", error=f"File not found: {path}")
except Exception as e:
    SUBMIT(status="error", error=f"{type(e).__name__}: {e}")
"""
        return _execute_submit_ctx(ctx, code, variables={"path": resolved_path})

    def memory_write(path: str, content: str) -> dict[str, Any]:
        """Write content to a file in persistent memory (Modal Volume)."""
        resolved_path, error = _resolve_path_or_error(
            path=path,
            default_root="/data/memory",
        )
        if error is not None:
            return error

        code = """
import os
import subprocess
try:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w", encoding="utf-8") as f:
        f.write(content)
    # Sync filesystem to volume if supported (optional but good for safety)
    try:
        os.sync()
    except AttributeError:
        pass
    sync_rc = 0
    try:
        proc = subprocess.run(["sync", "/data"], check=False, capture_output=True)
        sync_rc = int(proc.returncode)
    except Exception:
        sync_rc = -1
    SUBMIT(status="ok", path=path, chars=len(content), sync_rc=sync_rc)
except Exception as e:
    SUBMIT(status="error", error=f"{type(e).__name__}: {e}")
"""
        # Note: Modal volumes are eventually consistent, but os.sync() helps.
        # The Interpreter also exposes a .commit() method if needed on the host side.
        result = _execute_submit_ctx(
            ctx,
            code,
            variables={"path": resolved_path, "content": content},
        )

        # Trigger explicit commit on the host side for immediate persistence
        if result.get("status") == "ok":
            _commit_volume_best_effort(ctx)
        return result

    def write_to_file(path: str, content: str, append: bool = False) -> dict[str, Any]:
        """Write/append text to a file in Modal Volume with safe path handling."""
        if not append:
            return memory_write(path=path, content=content)

        resolved_path, error = _resolve_path_or_error(
            path=path,
            default_root="/data/memory",
        )
        if error is not None:
            return error

        code = """
import os
import subprocess
try:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "a", encoding="utf-8") as f:
        f.write(content)
    try:
        os.sync()
    except AttributeError:
        pass
    sync_rc = 0
    try:
        proc = subprocess.run(["sync", "/data"], check=False, capture_output=True)
        sync_rc = int(proc.returncode)
    except Exception:
        sync_rc = -1
    SUBMIT(status="ok", path=path, chars=len(content), mode="append", sync_rc=sync_rc)
except Exception as e:
    SUBMIT(status="error", error=f"{type(e).__name__}: {e}")
"""
        result = _execute_submit_ctx(
            ctx,
            code,
            variables={"path": resolved_path, "content": content},
        )
        if result.get("status") == "ok":
            _commit_volume_best_effort(ctx)
        return result

    def edit_core_memory(
        section: str,
        content: str,
        mode: str = "append",
    ) -> dict[str, Any]:
        """Edit core memory via append/replace operations with validation."""
        mode_norm = mode.strip().lower()
        if mode_norm not in {"append", "replace"}:
            return {
                "status": "error",
                "error": "mode must be one of: append, replace",
            }

        message = (
            ctx.agent.core_memory_append(section, content)
            if mode_norm == "append"
            else ctx.agent.core_memory_replace(section, content)
        )
        if message.startswith("Error:"):
            return {"status": "error", "error": message}

        return {
            "status": "ok",
            "section": section,
            "mode": mode_norm,
            "message": message,
            "chars": len(ctx.agent._core_memory.get(section, "")),
        }

    def memory_list(path: str = ".") -> dict[str, Any]:
        """List files and directories in persistent memory."""
        try:
            if path.strip() in {"", ".", "./"}:
                resolved_path = "/data/memory"
            else:
                resolved_path, error = _resolve_path_or_error(
                    path=path,
                    default_root="/data/memory",
                )
                if error is not None:
                    return error
        except AttributeError:
            resolved_path = "/data/memory"

        _reload_volume_best_effort(ctx)

        code = """
import os
try:
    items = []
    for name in os.listdir(path):
        full = os.path.join(path, name)
        kind = "dir" if os.path.isdir(full) else "file"
        items.append({"name": name, "type": kind})
    SUBMIT(status="ok", path=path, items=items, count=len(items))
except Exception as e:
    SUBMIT(status="error", error=f"{type(e).__name__}: {e}")
"""
        return _execute_submit_ctx(ctx, code, variables={"path": resolved_path})

    # -- Assemble tool list --------------------------------------------------

    from dspy import Tool

    # Sandbox editing tools
    tools.append(
        Tool(
            edit_file,
            name="edit_file",
            desc="Robustly edit a file by finding and replacing a unique text snippet",
        )
    )

    # Buffer & volume tools
    tools.extend(
        [
            Tool(
                read_buffer,
                name="read_buffer",
                desc="Read the full contents of a sandbox buffer",
            ),
            Tool(
                clear_buffer,
                name="clear_buffer",
                desc="Clear one sandbox buffer (or all buffers when name is empty)",
            ),
            Tool(
                save_buffer_to_volume,
                name="save_buffer_to_volume",
                desc="Persist a sandbox buffer to Modal Volume storage as JSON",
            ),
            Tool(
                load_text_from_volume,
                name="load_text_from_volume",
                desc="Load text from Modal Volume into host-side document memory",
            ),
            Tool(
                process_document,
                name="process_document",
                desc="Load a document from Modal Volume and register it for analysis",
            ),
        ]
    )

    # Persistent memory tools
    tools.extend(
        [
            Tool(
                memory_read,
                name="memory_read",
                desc="Read a file from persistent memory (Modal Volume)",
            ),
            Tool(
                memory_write,
                name="memory_write",
                desc="Write content to a file in persistent memory (Modal Volume)",
            ),
            Tool(
                write_to_file,
                name="write_to_file",
                desc="Write or append text to a file in persistent memory (Modal Volume)",
            ),
            Tool(
                edit_core_memory,
                name="edit_core_memory",
                desc="Edit core memory blocks using append or replace mode",
            ),
            Tool(
                memory_list,
                name="memory_list",
                desc="List files and directories in persistent memory",
            ),
        ]
    )

    return tools



"""Base and shared Pydantic schemas for the FastAPI layer."""

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class CamelCaseModel(BaseModel):
    """Base model that automatically aliases attributes from snake_case to camelCase.

    This ensures that Python snake_case variables serialize correctly for
    the frontend's expectations (e.g., created_at -> createdAt).
    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )



"""Auth domain types for server identity normalization."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class NormalizedIdentity:
    """Normalized identity extracted from auth claims."""

    tenant_claim: str
    user_claim: str
    email: str | None = None
    name: str | None = None
    raw_claims: dict[str, Any] | None = None



"""Header-based chunking strategy for markdown/structured text."""

from __future__ import annotations

import re


def chunk_by_headers(
    text: str,
    pattern: str = r"^#{1,3} ",
    flags: int = re.MULTILINE,
) -> list[dict]:
    """Split text by header boundaries (markdown-style).

    Splits a document at lines matching the given header pattern.
    Each chunk includes the header line and all content until the
    next header or end of document.

    Args:
        text: The text to split.
        pattern: Regex pattern matching header lines.
            Default: ``r"^#{1,3} "`` (markdown H1-H3).
        flags: Regex flags. Default: ``re.MULTILINE``.

    Returns:
        List of dicts with keys:
            - ``header``: The header line text (or "" for preamble)
            - ``content``: The content under that header
            - ``start_pos``: Character offset in original text

    Example:
        >>> text = "# Intro\\nHello\\n## Details\\nWorld"
        >>> chunks = chunk_by_headers(text)
        >>> chunks[0]["header"]
        '# Intro'
    """
    if not text:
        return []

    compiled = re.compile(pattern, flags)
    matches = list(compiled.finditer(text))

    if not matches:
        return [{"header": "", "content": text.strip(), "start_pos": 0}]

    chunks: list[dict] = []

    # Preamble before first header
    if matches[0].start() > 0:
        preamble = text[: matches[0].start()].strip()
        if preamble:
            chunks.append({"header": "", "content": preamble, "start_pos": 0})

    for i, match in enumerate(matches):
        # Get end position (start of next header or end of text)
        end = matches[i + 1].start() if i + 1 < len(matches) else len(text)
        section = text[match.start() : end]

        # Split header from content
        newline_pos = section.find("\n")
        if newline_pos == -1:
            header = section.strip()
            content = ""
        else:
            header = section[:newline_pos].strip()
            content = section[newline_pos + 1 :].strip()

        chunks.append(
            {"header": header, "content": content, "start_pos": match.start()}
        )

    return chunks



"""Sanitization helpers for analytics payloads."""

from __future__ import annotations

import re
from typing import Any


_SENSITIVE_PATTERNS: tuple[tuple[re.Pattern[str], str], ...] = (
    (re.compile(r"sk-[A-Za-z0-9_-]{8,}"), "sk-***REDACTED***"),
    (
        re.compile(r"(Authorization\s*:\s*Bearer\s+)[^\s]+", re.IGNORECASE),
        r"\1***REDACTED***",
    ),
    (
        re.compile(
            r"((?:api[_-]?key|token|secret|password)\s*[=:]\s*)(?:\"[^\"]*\"|'[^']*'|[^\s,}\]]+)",
            re.IGNORECASE,
        ),
        r"\1***REDACTED***",
    ),
)


def redact_sensitive(text: str) -> str:
    """Replace API keys and other sensitive tokens with redaction markers."""
    redacted = text
    for pattern, replacement in _SENSITIVE_PATTERNS:
        redacted = pattern.sub(replacement, redacted)
    return redacted


def truncate_text(text: str, max_chars: int) -> str:
    """Truncate text with a deterministic suffix describing omitted length."""
    if max_chars <= 0:
        return ""
    if len(text) <= max_chars:
        return text
    omitted = len(text) - max_chars
    return f"{text[:max_chars]}...[truncated, {omitted} more chars]"


def to_safe_text(value: Any) -> str:
    """Coerce arbitrary values to a stable string for analytics emission."""
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    try:
        return str(value)
    except Exception:
        return "<unprintable>"


def sanitize_text(text: str, *, redact: bool, truncation_chars: int) -> str:
    """Apply redaction and truncation in the correct order."""
    candidate = redact_sensitive(text) if redact else text
    return truncate_text(candidate, truncation_chars)



"""Router for session state management."""

from fastapi import APIRouter

from fleet_rlm.server.deps import ServerStateDep
from fleet_rlm.server.schemas.core import SessionStateResponse, SessionStateSummary

router = APIRouter(prefix="/sessions", tags=["sessions"])


@router.get("/state", response_model=SessionStateResponse)
async def list_session_state(state: ServerStateDep) -> SessionStateResponse:
    """Return lightweight summaries of active/restored in-memory session state."""
    summaries: list[SessionStateSummary] = []
    for key, payload in state.sessions.items():
        manifest = payload.get("manifest", {}) if isinstance(payload, dict) else {}
        session = payload.get("session", {}) if isinstance(payload, dict) else {}
        session_state = session.get("state", {}) if isinstance(session, dict) else {}
        history = (
            session_state.get("history", []) if isinstance(session_state, dict) else []
        )
        documents = (
            session_state.get("documents", {})
            if isinstance(session_state, dict)
            else {}
        )
        memory = manifest.get("memory", []) if isinstance(manifest, dict) else []
        logs = manifest.get("logs", []) if isinstance(manifest, dict) else []
        artifacts = manifest.get("artifacts", []) if isinstance(manifest, dict) else []
        metadata = manifest.get("metadata", {}) if isinstance(manifest, dict) else {}
        summaries.append(
            SessionStateSummary(
                key=key,
                workspace_id=str(payload.get("workspace_id", "default")),
                user_id=str(payload.get("user_id", "anonymous")),
                session_id=payload.get("session_id"),
                history_turns=len(history) if isinstance(history, list) else 0,
                document_count=len(documents) if isinstance(documents, dict) else 0,
                memory_count=len(memory) if isinstance(memory, list) else 0,
                log_count=len(logs) if isinstance(logs, list) else 0,
                artifact_count=len(artifacts) if isinstance(artifacts, list) else 0,
                updated_at=metadata.get("updated_at")
                if isinstance(metadata, dict)
                else None,
            )
        )
    return SessionStateResponse(ok=True, sessions=summaries)



"""Configuration models for PostHog LLM analytics."""

from __future__ import annotations

import os

from pydantic import BaseModel, Field

from fleet_rlm._env_utils import env_bool as _env_bool

PROJECT_POSTHOG_DEFAULT_HOST = "https://eu.i.posthog.com"
PROJECT_POSTHOG_DEFAULT_API_KEY: str | None = None


class PostHogConfig(BaseModel):
    """PostHog LLM analytics configuration."""

    enabled: bool = Field(default=False)
    api_key: str | None = Field(default=None)
    host: str = Field(default=PROJECT_POSTHOG_DEFAULT_HOST)
    flush_interval: float = Field(default=10.0)
    flush_at: int = Field(default=10)
    enable_dspy_optimization: bool = Field(default=False)
    input_truncation_chars: int = Field(default=10000)
    output_truncation_chars: int = Field(default=5000)
    redact_sensitive: bool = Field(default=True)

    @classmethod
    def from_env(cls) -> "PostHogConfig":
        """Load analytics configuration from environment variables."""
        api_key = (
            os.getenv("POSTHOG_API_KEY") or ""
        ).strip() or PROJECT_POSTHOG_DEFAULT_API_KEY
        host = (os.getenv("POSTHOG_HOST") or "").strip() or PROJECT_POSTHOG_DEFAULT_HOST
        enabled_raw = os.getenv("POSTHOG_ENABLED")
        return cls(
            enabled=_env_bool(enabled_raw, default=bool(api_key)),
            api_key=api_key,
            host=host,
            flush_interval=float(os.getenv("POSTHOG_FLUSH_INTERVAL", "10.0")),
            flush_at=max(1, int(os.getenv("POSTHOG_FLUSH_AT", "10"))),
            enable_dspy_optimization=_env_bool(
                os.getenv("POSTHOG_ENABLE_DSPY_OPTIMIZATION"), default=False
            ),
            input_truncation_chars=max(
                1, int(os.getenv("POSTHOG_INPUT_TRUNCATION", "10000"))
            ),
            output_truncation_chars=max(
                1, int(os.getenv("POSTHOG_OUTPUT_TRUNCATION", "5000"))
            ),
            redact_sensitive=_env_bool(
                os.getenv("POSTHOG_REDACT_SENSITIVE"), default=True
            ),
        )



"""Sandbox driver for Modal code execution via JSON protocol.

This module provides a long-lived JSON protocol driver that runs inside a Modal
sandbox. It receives code execution commands via stdin, executes them in a
controlled environment, and returns results via stdout.

The driver supports:
    - Stateful code execution (globals persist across calls)
    - Tool registration and invocation
    - Output capture (stdout/stderr)
    - Structured final output via SUBMIT function
    - Final variable convention (see below)

Protocol:
    Input (JSON, one per line):
        {
            "code": "python code to execute",
            "variables": {"var_name": value},  // Optional
            "tool_names": ["tool1", "tool2"],  // Optional
            "output_names": ["result1", "result2"],  // Optional
            "execution_profile": "ROOT_INTERLOCUTOR|RLM_DELEGATE|MAINTENANCE"  // Optional
        }

    Output (JSON, one per line):
        {
            "stdout": "captured stdout",
            "stderr": "captured stderr",
            "final": {...}  // Structured output from SUBMIT or Final
        }

    Tool calls (output):
        {"tool_call": {"name": "tool_name", "args": [...], "kwargs": {...}}}

    Tool responses (input):
        {"tool_result": ...} or {"tool_error": "error message"}

Final Variable Convention:
    As described in the RLM paper (Section 2), code executed in the REPL can
    signal completion by setting a variable named ``Final``. When the driver
    detects that ``Final`` has been set in the globals after code execution,
    it automatically returns the value of ``Final`` as the structured output
    and stops further iteration.

    This provides a natural way for LLM-generated code to indicate completion:

        >>> analysis = process_document(text)
        >>> Final = {"result": analysis, "status": "complete"}

    The driver will detect ``Final``, return its value, and terminate the
    session. If ``Final`` is not set, execution continues normally (backwards
    compatible with SUBMIT-based workflows).
"""

from __future__ import annotations

from .driver_factories import FinalOutput  # noqa: F401 — public re-export


def sandbox_driver() -> None:
    """Run the long-lived JSON protocol driver for Modal Sandbox execution.

    This function runs an infinite loop reading JSON commands from stdin,
    executing Python code, and writing results to stdout. It maintains
    state across executions through sandbox_globals.

    The driver provides these built-in capabilities to executed code:
        - SUBMIT(): Function to return structured final output
        - Final variable: Set a variable named ``Final`` to return structured
          output (alternative to SUBMIT, as per RLM paper Section 2)
        - llm_query(prompt): Query a sub-LLM for semantic analysis (via tool_call)
        - llm_query_batched(prompts): Query multiple prompts concurrently (via tool_call)
        - Tool functions: Dynamically registered based on tool_names in commands

    The loop terminates on EOFError (stdin closed) or when ``Final`` is set
    (if the caller stops sending commands after receiving a Final response).
    """
    # Keep these imports inside the function so the source extracted by
    # ``inspect.getsource(sandbox_driver)`` is self-contained when executed
    # in the Modal sandbox process.
    import json
    import sys
    from contextlib import redirect_stderr, redirect_stdout
    from io import StringIO
    from typing import Any, Callable, cast

    try:
        from fleet_rlm.core.driver_factories import (
            FinalOutput,
            inject_sandbox_helpers,
            make_llm_query,
            make_llm_query_batched,
            make_send,
            make_submit,
            make_tool_call,
            register_tools,
            wrap_helper,
        )
        from fleet_rlm.core.sandbox_tools import (
            add_buffer,
            chunk_by_headers,
            chunk_by_json_keys,
            chunk_by_size,
            chunk_by_timestamps,
            clear_buffer,
            get_buffer,
            grep,
            peek,
            reset_buffers,
        )
        from fleet_rlm.core.session_history import (
            get_last_execution,
            get_session_history,
            log_execution,
            reset_session_history,
        )
        from fleet_rlm.core.volume_tools import (
            load_from_volume,
            save_to_volume,
            workspace_append,
            workspace_list,
            workspace_read,
            workspace_write,
        )
    except ModuleNotFoundError:
        # In Modal, interpreter may execute a bundled script that already
        # defines these symbols in globals without an installed fleet_rlm package.
        # Access them from globals() instead.
        g: dict[str, Any] = globals()
        FinalOutput = cast(Any, g.get("FinalOutput"))
        inject_sandbox_helpers = cast(Any, g.get("inject_sandbox_helpers"))
        make_llm_query = cast(Any, g.get("make_llm_query"))
        make_llm_query_batched = cast(Any, g.get("make_llm_query_batched"))
        make_send = cast(Any, g.get("make_send"))
        make_submit = cast(Any, g.get("make_submit"))
        make_tool_call = cast(Any, g.get("make_tool_call"))
        register_tools = cast(Any, g.get("register_tools"))
        wrap_helper = cast(Any, g.get("wrap_helper"))
        add_buffer = cast(Any, g.get("add_buffer"))
        chunk_by_headers = cast(Any, g.get("chunk_by_headers"))
        chunk_by_json_keys = cast(Any, g.get("chunk_by_json_keys"))
        chunk_by_size = cast(Any, g.get("chunk_by_size"))
        chunk_by_timestamps = cast(Any, g.get("chunk_by_timestamps"))
        clear_buffer = cast(Any, g.get("clear_buffer"))
        get_buffer = cast(Any, g.get("get_buffer"))
        grep = cast(Any, g.get("grep"))
        peek = cast(Any, g.get("peek"))
        reset_buffers = cast(Any, g.get("reset_buffers"))
        get_last_execution = cast(Any, g.get("get_last_execution"))
        get_session_history = cast(Any, g.get("get_session_history"))
        log_execution = cast(Any, g.get("log_execution"))
        reset_session_history = cast(Any, g.get("reset_session_history"))
        load_from_volume = cast(Any, g.get("load_from_volume"))
        save_to_volume = cast(Any, g.get("save_to_volume"))
        workspace_append = cast(Any, g.get("workspace_append"))
        workspace_list = cast(Any, g.get("workspace_list"))
        workspace_read = cast(Any, g.get("workspace_read"))
        workspace_write = cast(Any, g.get("workspace_write"))

    # Reset module-level state for fresh start (each driver instance is independent)
    reset_session_history()
    reset_buffers()

    # Persistent globals that survive across code execution calls
    sandbox_globals: dict[str, Any] = {}
    proto_out = sys.__stdout__

    # Use list for mutable reference in closure
    current_execution_profile = ["RLM_DELEGATE"]
    _dynamic_tool_names: set[str] = set()

    output_names: list[str] = []

    # Create protocol functions
    _send = make_send(proto_out)
    _tool_call = make_tool_call(_send)

    # Create wrapped helper injector
    def _wrap(fn: Callable) -> Callable:
        return wrap_helper(fn, current_execution_profile)

    # Inject sandbox helpers into globals
    inject_sandbox_helpers(
        sandbox_globals,
        _wrap,
        {
            "peek": peek,
            "grep": grep,
            "chunk_by_size": chunk_by_size,
            "chunk_by_headers": chunk_by_headers,
            "chunk_by_timestamps": chunk_by_timestamps,
            "chunk_by_json_keys": chunk_by_json_keys,
            "add_buffer": add_buffer,
            "get_buffer": get_buffer,
            "clear_buffer": clear_buffer,
        },
        {
            "save_to_volume": save_to_volume,
            "load_from_volume": load_from_volume,
            "workspace_write": workspace_write,
            "workspace_read": workspace_read,
            "workspace_list": workspace_list,
            "workspace_append": workspace_append,
        },
        {
            "log_execution": log_execution,
            "get_session_history": get_session_history,
            "get_last_execution": get_last_execution,
        },
    )

    # Main execution loop
    while True:
        try:
            line = input()
        except EOFError:
            break

        try:
            command = json.loads(line)
        except json.JSONDecodeError as exc:
            _send(
                {"stdout": "", "stderr": f"[Error] Invalid JSON: {exc}", "final": None}
            )
            continue

        code = command.get("code")
        variables = command.get("variables", {}) or {}
        tool_names = list(command.get("tool_names", []) or [])
        output_names = list(command.get("output_names", []) or [])
        execution_profile = str(
            command.get("execution_profile", "RLM_DELEGATE")
        ).strip()
        if execution_profile not in {
            "ROOT_INTERLOCUTOR",
            "RLM_DELEGATE",
            "MAINTENANCE",
        }:
            execution_profile = "RLM_DELEGATE"
        current_execution_profile[0] = execution_profile

        if code is None:
            _send({"stdout": "", "stderr": "[Error] No code provided", "final": None})
            continue

        # Create SUBMIT for this execution
        SUBMIT = make_submit(output_names)
        sandbox_globals["SUBMIT"] = SUBMIT

        # Create LLM query functions
        llm_query = make_llm_query(_tool_call)
        llm_query_batched = make_llm_query_batched(_tool_call)
        sandbox_globals["llm_query"] = llm_query
        sandbox_globals["llm_query_batched"] = llm_query_batched

        sandbox_globals.update(variables)
        register_tools(
            tool_names,
            sandbox_globals,
            _dynamic_tool_names,
            _tool_call,
            current_execution_profile,
        )

        stdout_io = StringIO()
        stderr_io = StringIO()
        final_obj = None

        had_exec_error = False
        with redirect_stdout(stdout_io), redirect_stderr(stderr_io):
            try:
                exec(code, sandbox_globals)
            except FinalOutput as exc:
                final_obj = exc.args[0] if exc.args else None
            except Exception as exc:  # pragma: no cover
                had_exec_error = True
                print(f"[Error] {type(exc).__name__}: {exc}", file=sys.stderr)

        # Final Variable Convention: Check if 'Final' was set in globals.
        # Always clear it after execution to prevent stale values leaking into
        # later commands in this stateful session.
        if had_exec_error:
            sandbox_globals.pop("Final", None)
        else:
            _missing = object()
            final_from_var = sandbox_globals.pop("Final", _missing)
            if final_obj is None and final_from_var is not _missing:
                final_obj = final_from_var

        result = {
            "stdout": stdout_io.getvalue(),
            "stderr": stderr_io.getvalue(),
            "final": final_obj,
        }

        # Log execution for session history tracking
        log_execution(code, result, {"had_error": had_exec_error})

        _send(result)



"""PostHog analytics integration for DSPy LM calls."""

from __future__ import annotations

import dspy

from .client import flush_posthog_client, shutdown_posthog_client
from .config import PostHogConfig
from .posthog_callback import PostHogLLMCallback


__all__ = [
    "PostHogConfig",
    "PostHogLLMCallback",
    "configure_analytics",
    "flush_posthog_client",
    "shutdown_posthog_client",
]


def _get_existing_callback() -> PostHogLLMCallback | None:
    callbacks = list(getattr(dspy.settings, "callbacks", []) or [])
    for callback in callbacks:
        if isinstance(callback, PostHogLLMCallback):
            return callback
    return None


def configure_analytics(
    *,
    api_key: str | None = None,
    host: str = "https://us.i.posthog.com",
    distinct_id: str | None = None,
    enabled: bool | None = None,
) -> PostHogLLMCallback | None:
    """Configure and register PostHog analytics callback with DSPy."""
    base = PostHogConfig.from_env()
    resolved_host = (
        base.host if host == "https://us.i.posthog.com" and base.host else host
    )
    config = PostHogConfig(
        enabled=base.enabled if enabled is None else enabled,
        api_key=api_key if api_key is not None else base.api_key,
        host=resolved_host,
        flush_interval=base.flush_interval,
        flush_at=base.flush_at,
        enable_dspy_optimization=base.enable_dspy_optimization,
        input_truncation_chars=base.input_truncation_chars,
        output_truncation_chars=base.output_truncation_chars,
        redact_sensitive=base.redact_sensitive,
    )

    if not config.enabled or not config.api_key:
        return None

    existing = _get_existing_callback()
    if existing is not None:
        # Keep registration idempotent and avoid callback duplication.
        return existing

    callback = PostHogLLMCallback(config, distinct_id=distinct_id)
    existing_callbacks = list(getattr(dspy.settings, "callbacks", []) or [])
    dspy.configure(callbacks=[*existing_callbacks, callback])
    return callback



"""Structured execution-event models and streaming helpers.

This module powers the dedicated ``/ws/execution`` event stream consumed by
Artifact Canvas-style visualizations.
"""

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from threading import RLock
from typing import Any, Literal, TYPE_CHECKING

from fastapi import WebSocket
from pydantic import BaseModel

from .sanitizer import sanitize_event_payload, summarize_code_for_event

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .step_builder import ExecutionStepBuilder

ExecutionStepType = Literal["llm", "tool", "repl", "memory", "output"]
ExecutionActorKind = Literal["root_rlm", "sub_agent", "delegate", "unknown"]
ExecutionEventType = Literal[
    "execution_started",
    "execution_step",
    "execution_completed",
]


class ExecutionStep(BaseModel):
    """Single execution graph node/edge payload."""

    id: str
    parent_id: str | None = None
    type: ExecutionStepType
    label: str
    depth: int | None = None
    actor_kind: ExecutionActorKind | None = None
    actor_id: str | None = None
    lane_key: str | None = None
    input: Any | None = None
    output: Any | None = None
    timestamp: float


class ExecutionEvent(BaseModel):
    """Top-level event envelope emitted over ``/ws/execution``."""

    type: ExecutionEventType
    run_id: str
    workspace_id: str
    user_id: str
    session_id: str
    step: ExecutionStep | None = None


class ExecutionSubscription(BaseModel):
    """Required identity filter for execution-stream subscriptions."""

    workspace_id: str
    user_id: str
    session_id: str

    def matches(self, event: ExecutionEvent) -> bool:
        return (
            self.workspace_id == event.workspace_id
            and self.user_id == event.user_id
            and self.session_id == event.session_id
        )


class ExecutionEventEmitter:
    """Broadcast ``ExecutionEvent`` payloads to matching websocket subscribers."""

    @dataclass(slots=True)
    class _ConnectionState:
        subscription: ExecutionSubscription
        queue: asyncio.Queue[dict[str, Any] | None]
        sender_task: asyncio.Task[None]
        dropped_events: int = 0

    def __init__(
        self,
        *,
        max_queue: int = 256,
        drop_policy: Literal["drop_oldest", "drop_newest"] = "drop_oldest",
    ) -> None:
        self._max_queue = max(1, int(max_queue))
        self._drop_policy = drop_policy
        self._connections: dict[WebSocket, ExecutionEventEmitter._ConnectionState] = {}
        self._lock = RLock()

    async def connect(
        self, websocket: WebSocket, subscription: ExecutionSubscription
    ) -> None:
        await websocket.accept()
        queue: asyncio.Queue[dict[str, Any] | None] = asyncio.Queue(
            maxsize=self._max_queue
        )
        sender_task = asyncio.create_task(self._sender_loop(websocket))
        state = self._ConnectionState(
            subscription=subscription,
            queue=queue,
            sender_task=sender_task,
        )
        with self._lock:
            self._connections[websocket] = state

    async def disconnect(self, websocket: WebSocket) -> None:
        state: ExecutionEventEmitter._ConnectionState | None = None
        with self._lock:
            state = self._connections.pop(websocket, None)
        if state is None:
            return

        try:
            state.queue.put_nowait(None)
        except asyncio.QueueFull:
            try:
                _ = state.queue.get_nowait()
                state.queue.put_nowait(None)
            except (asyncio.QueueEmpty, asyncio.QueueFull):
                logger.debug(
                    "Ignoring execution queue race during disconnect teardown",
                )

        current_task = asyncio.current_task()
        if state.sender_task is not current_task:
            state.sender_task.cancel()
            try:
                await state.sender_task
            except asyncio.CancelledError:
                # Normal outcome: task was cancelled during disconnect teardown.
                pass
            except Exception:
                logger.exception(
                    "Unexpected error while awaiting sender_task during disconnect teardown",
                )

    async def _sender_loop(self, websocket: WebSocket) -> None:
        state: ExecutionEventEmitter._ConnectionState | None = None
        while True:
            with self._lock:
                state = self._connections.get(websocket)
            if state is None:
                return

            payload = await state.queue.get()
            if payload is None:
                break
            try:
                await websocket.send_json(payload)
            except Exception:
                break

        await self.disconnect(websocket)

    def _enqueue_payload(
        self,
        state: _ConnectionState,
        payload: dict[str, Any],
    ) -> None:
        try:
            state.queue.put_nowait(payload)
            return
        except asyncio.QueueFull:
            pass

        if self._drop_policy == "drop_newest":
            state.dropped_events += 1
            return

        # Default drop policy: keep latest signal and evict the oldest entry.
        try:
            _ = state.queue.get_nowait()
            state.dropped_events += 1
        except asyncio.QueueEmpty:
            state.dropped_events += 1
            return

        try:
            state.queue.put_nowait(payload)
        except asyncio.QueueFull:
            state.dropped_events += 1

    async def emit(self, event: ExecutionEvent) -> None:
        payload = event.model_dump(mode="json")
        with self._lock:
            targets = [
                state
                for state in self._connections.values()
                if state.subscription.matches(event)
            ]
        for state in targets:
            self._enqueue_payload(state, payload)

    def dropped_event_count(self) -> int:
        with self._lock:
            return sum(state.dropped_events for state in self._connections.values())


# Late import avoids a circular dependency during module initialization while
# still providing a concrete symbol for static analyzers and re-export users.
from .step_builder import ExecutionStepBuilder  # noqa: E402


__all__ = [
    "ExecutionEvent",
    "ExecutionEventEmitter",
    "ExecutionActorKind",
    "ExecutionEventType",
    "ExecutionStep",
    "ExecutionStepBuilder",
    "ExecutionStepType",
    "ExecutionSubscription",
    "sanitize_event_payload",
    "summarize_code_for_event",
]



"""Pydantic models used by the interactive coding CLI runtime."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Literal

from pydantic import BaseModel, Field

TraceMode = Literal["compact", "verbose", "off"]


class ProfileConfig(BaseModel):
    """Persistent CLI profile for code-chat sessions."""

    name: str
    docs_path: str | None = None
    secret_name: str = "LITELLM"
    volume_name: str | None = None
    timeout: int = 900
    react_max_iters: int = 10
    rlm_max_iterations: int = 30
    rlm_max_llm_calls: int = 50
    trace: bool = False
    stream: bool = True


class SessionConfig(BaseModel):
    """Runtime config for an interactive session."""

    profile_name: str = "default"
    docs_path: str | None = None
    secret_name: str = "LITELLM"
    volume_name: str | None = None
    timeout: int = 900
    react_max_iters: int = 10
    rlm_max_iterations: int = 30
    rlm_max_llm_calls: int = 50
    trace: bool = True
    trace_mode: TraceMode = "compact"
    stream: bool = True
    stream_refresh_ms: int = 40


class CommandResult(BaseModel):
    """Structured command execution result for UI rendering."""

    ok: bool = True
    kind: Literal["info", "error", "assistant", "trace", "data"] = "info"
    message: str = ""
    payload: dict | list | str | None = None


class TranscriptEvent(BaseModel):
    """JSONL event persisted for a chat session transcript."""

    role: Literal["user", "assistant", "system", "trace", "status"]
    content: str = Field(default="")
    payload: dict | None = None


StreamEventKind = Literal[
    "assistant_token",
    "status",
    "reasoning_step",
    "tool_call",
    "tool_result",
    "trajectory_step",
    "plan_update",
    "rlm_executing",
    "memory_update",
    "final",
    "error",
    "cancelled",
]


@dataclass(slots=True)
class StreamEvent:
    """Streaming event emitted from agent runtime to UI surfaces."""

    kind: StreamEventKind
    text: str = ""
    payload: dict[str, Any] = field(default_factory=dict)
    timestamp: datetime = field(default_factory=lambda: datetime.now(timezone.utc))
    # Control signal: flush any batched tokens before emitting this event
    flush_tokens: bool = False


@dataclass(slots=True)
class TurnState:
    """Accumulates stream events into a render-ready turn state."""

    assistant_tokens: list[str] = field(default_factory=list)
    transcript_text: str = ""
    reasoning_lines: list[str] = field(default_factory=list)
    tool_timeline: list[str] = field(default_factory=list)
    status_lines: list[str] = field(default_factory=list)
    stream_chunks: list[str] = field(default_factory=list)
    thought_chunks: list[str] = field(default_factory=list)
    status_messages: list[str] = field(default_factory=list)
    trajectory: dict[str, Any] = field(default_factory=dict)
    final_text: str = ""
    final_reasoning: str = ""
    history_turns: int = 0
    token_count: int = 0
    cancelled: bool = False
    errored: bool = False
    done: bool = False
    error_message: str = ""

    def apply(self, event: StreamEvent) -> None:
        """Apply one event to state in a deterministic way."""
        if event.kind == "assistant_token":
            token = event.text
            self.assistant_tokens.append(token)
            self.stream_chunks.append(token)
            self.token_count += 1
            self.transcript_text = "".join(self.assistant_tokens)
            return

        if event.kind == "status":
            if event.text:
                self.status_lines.append(event.text)
                self.status_messages.append(event.text)
                self.reasoning_lines.append(event.text)
            return

        if event.kind == "reasoning_step":
            if event.text:
                self.reasoning_lines.append(event.text)
                self.thought_chunks.append(event.text)
            return

        if event.kind == "tool_call":
            if event.text:
                self.tool_timeline.append(event.text)
            return

        if event.kind == "tool_result":
            if event.text:
                self.tool_timeline.append(event.text)
            return

        if event.kind == "trajectory_step":
            step_data = event.payload.get("step_data", {})
            if step_data:
                current_steps = self.trajectory.get("steps", [])
                current_steps.append(step_data)
                self.trajectory["steps"] = current_steps
            return

        if event.kind in ("plan_update", "rlm_executing", "memory_update"):
            if event.text:
                # Maintain CLI backward compatibility by casting these as status/timeline noise
                self.status_lines.append(event.text)
                self.tool_timeline.append(event.text)
            return

        if event.kind == "final":
            final_text = event.text or self.transcript_text
            self.final_text = final_text
            self.transcript_text = final_text
            self.trajectory = dict(event.payload.get("trajectory", {}) or {})
            self.final_reasoning = event.payload.get("final_reasoning", "")
            self.history_turns = int(
                event.payload.get("history_turns", self.history_turns)
            )
            self.done = True
            return

        if event.kind == "cancelled":
            self.cancelled = True
            self.done = True
            cancelled_text = event.text or self.transcript_text
            self.final_text = cancelled_text
            self.transcript_text = cancelled_text
            self.history_turns = int(
                event.payload.get("history_turns", self.history_turns)
            )
            return

        if event.kind == "error":
            self.errored = True
            self.done = True
            self.error_message = event.text or "unknown error"
            self.history_turns = int(
                event.payload.get("history_turns", self.history_turns)
            )



"""UI and rendering functions for the terminal chat interface.

This module provides UI and rendering components extracted from terminal_chat.py.
All functions are stateless and take required parameters explicitly.
"""

from __future__ import annotations

import getpass
import json
import re
from pathlib import Path
from typing import Any

from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import input_dialog, radiolist_dialog
from prompt_toolkit.styles import Style
from rich.layout import Layout
from rich.panel import Panel
from rich.text import Text


class _FleetCompleter(Completer):
    """Slash + file mention completer used by the chat composer.

    This completer provides autocomplete for slash commands and file mentions
    (using @ prefix) in the terminal chat interface.
    """

    def __init__(
        self,
        command_specs: list[tuple[str, str]],
        command_dispatch_names: list[str] | None = None,
    ) -> None:
        """Initialize the completer.

        Args:
            command_specs: List of (command_name, summary) tuples.
            command_dispatch_names: List of command dispatch names for tool commands.
        """
        dispatch_names = command_dispatch_names or []
        self._slash_entries: list[tuple[str, str]] = sorted(
            command_specs
            + [(f"/{name}", "tool command") for name in sorted(dispatch_names)],
            key=lambda item: item[0],
        )

    def get_completions(self, document: Any, complete_event: Any):
        """Generate completions for the current input."""
        text = document.text_before_cursor

        if text.startswith("/"):
            token = text.split(maxsplit=1)[0]
            for command, summary in self._slash_entries:
                if command.startswith(token):
                    yield Completion(
                        command,
                        start_position=-len(token),
                        display=command,
                        display_meta=summary,
                    )
            if text.startswith("/settings "):
                sub = (
                    text.split(maxsplit=1)[1] if len(text.split(maxsplit=1)) > 1 else ""
                )
                for choice in ("llm", "model", "modal"):
                    if choice.startswith(sub):
                        yield Completion(
                            choice,
                            start_position=-len(sub),
                            display=choice,
                            display_meta="settings scope",
                        )
            return

        mention = re.search(r"@(\S*)$", text)
        if not mention:
            return
        prefix = mention.group(1)
        for candidate in _iter_mention_paths(prefix):
            yield Completion(
                candidate,
                start_position=-len(prefix),
                display=f"@{candidate}",
                display_meta="file path",
            )


def _history_path() -> Path:
    """Return the path to the chat history file.

    Returns:
        Path to ~/.fleet/history.txt
    """
    return Path.home() / ".fleet" / "history.txt"


def _badge(ok: bool) -> str:
    """Return a status badge string.

    Args:
        ok: Whether the status is OK.

    Returns:
        Rich markup string for the badge.
    """
    return "[green]OK[/]" if ok else "[yellow]WARN[/]"


def _mask_secret(value: str) -> str:
    """Mask a secret value for display.

    Args:
        value: The secret value to mask.

    Returns:
        Masked string showing only first 3 and last 2 characters.
    """
    if not value:
        return ""
    if len(value) <= 6:
        return "*" * len(value)
    return f"{value[:3]}...{value[-2:]}"


def _prompt_label() -> HTML:
    """Return the prompt label HTML.

    Returns:
        HTML formatted prompt label.
    """
    return HTML("<prompt>❯ </prompt>")


def _bottom_toolbar(*, is_processing: bool) -> HTML:
    """Return the bottom toolbar HTML.

    Args:
        is_processing: Whether the session is currently processing a request.

    Returns:
        HTML formatted toolbar text.
    """
    mode = "thinking" if is_processing else "idle"
    return HTML(
        "<trace>"
        " Type @ to mention files, / for commands, or ? for shortcuts "
        f" · mode={mode}"
        "</trace>"
    )


def _iter_mention_paths(prefix: str, *, limit: int = 40) -> list[str]:
    """Iterate over file paths matching a prefix for @ mention completion.

    Args:
        prefix: The path prefix to match.
        limit: Maximum number of suggestions to return.

    Returns:
        List of matching path suggestions.
    """
    base = Path.cwd()
    query = prefix.strip()
    prefix_dir = ""
    name_prefix = query
    root = base

    if query:
        as_path = Path(query)
        if as_path.is_absolute():
            parent = as_path.parent if as_path.parent.as_posix() else Path("/")
            root = parent if parent.exists() else Path("/")
            prefix_dir = (
                f"{as_path.parent.as_posix().rstrip('/')}/"
                if as_path.parent.as_posix()
                else ""
            )
            name_prefix = as_path.name
        elif "/" in query:
            maybe_dir, name_prefix = query.rsplit("/", 1)
            resolved = (base / maybe_dir).resolve()
            if resolved.exists() and resolved.is_dir():
                root = resolved
                prefix_dir = maybe_dir.rstrip("/") + "/"

    suggestions: list[str] = []
    try:
        entries = sorted(
            root.iterdir(), key=lambda path: (not path.is_dir(), path.name.lower())
        )
    except Exception:  # directory may not exist or be unreadable
        return suggestions

    if not query:
        suggestions.append(str(base))

    lowered = name_prefix.lower()
    for entry in entries:
        if lowered and not entry.name.lower().startswith(lowered):
            continue
        suffix = "/" if entry.is_dir() else ""
        suggestion = f"{prefix_dir}{entry.name}{suffix}"
        suggestions.append(suggestion)
        if len(suggestions) >= limit:
            break
    return suggestions


def _prompt_value(*, key: str, default: str, secret: bool) -> str:
    """Prompt the user for a value with optional secret masking.

    Args:
        key: The configuration key name.
        default: Default value to show.
        secret: Whether to mask the input.

    Returns:
        The user's input value.
    """
    shown_default = _mask_secret(default) if secret else default
    suffix = f" [{shown_default}]" if shown_default else ""
    try:
        if secret:
            raw = getpass.getpass(f"{key}{suffix}: ").strip()
        else:
            raw = input(f"{key}{suffix}: ").strip()
    except (EOFError, KeyboardInterrupt):  # pragma: no cover - interactive path
        return ""
    return raw


def _prompt_choice(
    prompt: str,
    choices: list[str],
    *,
    allow_freeform: bool,
) -> str:
    """Prompt the user to select from a list of choices.

    Args:
        prompt: The prompt text to display.
        choices: List of choices to present.
        allow_freeform: Whether to allow custom input.

    Returns:
        The selected choice or custom input.
    """
    try:
        values = [(str(index), choice) for index, choice in enumerate(choices, start=1)]
        if allow_freeform:
            values.append(("0", "Custom input"))
        picked = radiolist_dialog(
            title="Select option",
            text=prompt,
            values=values,
            style=_dialog_style(),
        ).run()
        if picked is None:
            return ""
        if picked == "0":
            custom = input_dialog(
                title="Custom input",
                text=prompt,
                style=_dialog_style(),
            ).run()
            return (custom or "").strip()
        if picked.isdigit():
            number = int(picked)
            if 1 <= number <= len(choices):
                return choices[number - 1]
    except (
        Exception
    ):  # prompt_toolkit dialog unavailable; fall back to plain print-based menu
        pass

    print(prompt)
    for index, choice in enumerate(choices, start=1):
        print(f"  {index}) {choice}")
    if allow_freeform:
        print("  0) custom input")
    while True:
        selection = input("Select option: ").strip()
        if selection.isdigit():
            number = int(selection)
            if 1 <= number <= len(choices):
                return choices[number - 1]
            if allow_freeform and number == 0:
                return input("Custom value: ").strip()
        if allow_freeform and selection:
            return selection
        print("Invalid selection.")


def _dialog_style() -> Style:
    """Return the dialog style for prompt_toolkit dialogs.

    Returns:
        Style object with dark theme colors.
    """
    return Style.from_dict(
        {
            "dialog": "bg:#101114",
            "dialog frame.label": "fg:#77d6ff bold",
            "dialog.body": "fg:#f0f3f7",
            "dialog shadow": "bg:#050607",
            "button.focused": "bg:#2d7ff9 #ffffff",
        }
    )


def _render_shell(
    *,
    console: Any,
    session_id: str,
    model: str,
    trace_mode: str,
    last_status: str,
    transcript: list[tuple[str, str]],
    is_processing: bool,
    draft_assistant: str = "",
) -> None:
    """Render the shell UI layout.

    Args:
        console: The Rich console instance.
        session_id: Current session identifier.
        model: Current model name.
        trace_mode: Current trace mode.
        last_status: Last status message.
        transcript: List of (role, content) transcript entries.
        is_processing: Whether currently processing a request.
        draft_assistant: Draft assistant response text.
    """
    header = Panel(
        Text.from_markup(
            f"[bold]fleet[/]  "
            f"[dim]session[/]={session_id}  "
            f"[dim]model[/]={model}  "
            f"[dim]trace[/]={trace_mode}  "
            f"[dim]status[/]={last_status}"
        ),
        border_style="cyan",
        padding=(0, 1),
    )

    body_text = Text()
    for role, content in transcript[-30:]:
        role_style = {
            "you": "bold white",
            "assistant": "bold cyan",
            "status": "dim",
            "tool": "magenta",
            "thinking": "dim",
            "warning": "yellow",
            "error": "red",
            "result": "blue",
        }.get(role, "white")
        body_text.append(f"{role}> ", style=role_style)
        body_text.append(content + "\n\n")

    if is_processing and draft_assistant:
        body_text.append("assistant> ", style="bold cyan")
        body_text.append(draft_assistant + "\n")

    transcript_panel = Panel(
        body_text if body_text.plain.strip() else Text("No messages yet.", style="dim"),
        border_style="bright_black",
        title="chat",
    )

    hint = (
        "Enter=send • Shift+Enter=newline • /=command palette • "
        "Ctrl+C=interrupt • /help=commands"
    )
    footer = Panel(Text(hint, style="dim"), border_style="bright_black")

    layout = Layout()
    layout.split_column(
        Layout(header, size=3),
        Layout(transcript_panel, ratio=1),
        Layout(footer, size=3),
    )

    console.clear()
    console.print(layout)


def _print_banner(
    *,
    console: Any,
    session_id: str,
    model: str,
    planner_ready: bool,
    workspace: Path,
) -> None:
    """Print the startup banner.

    Args:
        console: The Rich console instance.
        session_id: Current session identifier.
        model: Current model name.
        planner_ready: Whether the planner LM is configured.
        workspace: Current workspace path.
    """
    planner_text = "[green]ready[/]" if planner_ready else "[yellow]not configured[/]"
    content = (
        "[bold cyan]fleet[/]  [dim]Copilot-style mode[/]\n"
        "Describe a task to get started.\n\n"
        "Use [bold]/model[/], [bold]/settings[/], [bold]/status[/], and "
        "[bold]/[/] for the command palette.\n"
        f"[dim]session={session_id}  planner={planner_text}[/]"
    )
    console.print(Panel(content, title="fleet chat", border_style="cyan", expand=False))
    console.print(f"[dim]• workspace:[/] {workspace}    [dim]model:[/] {model}")


def _print_result_inline(
    *,
    console: Any,
    result: dict[str, Any],
    title: str,
) -> None:
    """Print a result as a panel.

    Args:
        console: The Rich console instance.
        result: Result dictionary to print.
        title: Title for the result.
    """
    rendered = json.dumps(result, indent=2, sort_keys=True, default=str)
    console.print(Panel(rendered, title=title, border_style="blue"))


# Role style mapping for transcript rendering
ROLE_STYLES: dict[str, str] = {
    "you": "bold white",
    "assistant": "bold cyan",
    "status": "dim",
    "tool": "magenta",
    "thinking": "dim",
    "warning": "yellow",
    "error": "red",
    "result": "blue",
}



"""Shared adapters for interpreter results and stateful response shaping."""

from __future__ import annotations

from typing import Any

from dspy.primitives.code_interpreter import FinalOutput


def final_output_value(result: Any) -> Any | None:
    """Return wrapped payload when result is FinalOutput, else None."""
    if isinstance(result, FinalOutput):
        return result.output
    return None


def final_output_dict(result: Any) -> dict[str, Any] | None:
    """Return output dict when interpreter result is a supported mapping payload."""
    output = final_output_value(result)
    if output is not None:
        if isinstance(output, dict):
            return output
        return None
    if isinstance(result, dict):
        return result
    return None


def extract_execute_error(result: Any) -> str | None:
    """Return stderr-like error string from interpreter string output."""
    if isinstance(result, str):
        stripped = result.strip()
        if "[Error]" in stripped:
            return stripped
    return None


def operation_error(filename: str, error: Any) -> dict[str, Any]:
    """Standardized error dictionary for workspace operations."""
    return {
        "status": "error",
        "filename": filename,
        "error": str(error),
    }


def operation_unexpected(filename: str) -> dict[str, Any]:
    """Standardized unexpected-result dictionary for workspace operations."""
    return operation_error(filename, "Unexpected result format")



"""Router module exports."""

from . import (
    auth,
    chat,
    health,
    runtime,
    sessions,
    ws,
)

__all__ = [
    "health",
    "auth",
    "chat",
    "ws",
    "sessions",
    "runtime",
]



"""SQLAlchemy models for fleet-rlm Neon Postgres persistence."""

from __future__ import annotations

import enum
import uuid
from datetime import datetime

from sqlalchemy import (
    BigInteger,
    Boolean,
    CheckConstraint,
    DateTime,
    Enum,
    ForeignKey,
    ForeignKeyConstraint,
    Index,
    Integer,
    SmallInteger,
    String,
    Text,
    UniqueConstraint,
    func,
    text,
)
from sqlalchemy.dialects.postgresql import ARRAY, JSONB, UUID
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    """Base class for all ORM models."""


def _pg_enum(enum_cls: type[enum.Enum], *, name: str) -> Enum:
    """Bind Python enums to existing Postgres enum values (not member names)."""
    return Enum(
        enum_cls,
        name=name,
        values_callable=lambda members: [member.value for member in members],
        validate_strings=True,
        native_enum=True,
    )


class TenantPlan(str, enum.Enum):
    FREE = "free"
    TEAM = "team"
    ENTERPRISE = "enterprise"


class TenantStatus(str, enum.Enum):
    ACTIVE = "active"
    SUSPENDED = "suspended"
    DELETED = "deleted"


class MembershipRole(str, enum.Enum):
    OWNER = "owner"
    ADMIN = "admin"
    MEMBER = "member"
    VIEWER = "viewer"


class SandboxProvider(str, enum.Enum):
    MODAL = "modal"
    DAYTONA = "daytona"
    ACA_JOBS = "aca_jobs"
    LOCAL = "local"


class SandboxSessionStatus(str, enum.Enum):
    ACTIVE = "active"
    ENDED = "ended"
    FAILED = "failed"


class RunStatus(str, enum.Enum):
    QUEUED = "queued"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"


class RunStepType(str, enum.Enum):
    TOOL_CALL = "tool_call"
    REPL_EXEC = "repl_exec"
    LLM_CALL = "llm_call"
    RETRIEVAL = "retrieval"
    GUARDRAIL = "guardrail"
    SUMMARY = "summary"
    MEMORY = "memory"
    OUTPUT = "output"
    STATUS = "status"


class JobType(str, enum.Enum):
    RUN_TASK = "run_task"
    MEMORY_COMPACTION = "memory_compaction"
    EVALUATION = "evaluation"
    MAINTENANCE = "maintenance"


class JobStatus(str, enum.Enum):
    QUEUED = "queued"
    LEASED = "leased"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    DEAD = "dead"


class MemoryScope(str, enum.Enum):
    USER = "user"
    TENANT = "tenant"
    RUN = "run"
    AGENT = "agent"


class MemoryKind(str, enum.Enum):
    NOTE = "note"
    SUMMARY = "summary"
    FACT = "fact"
    PREFERENCE = "preference"
    CONTEXT = "context"


class MemorySource(str, enum.Enum):
    USER_INPUT = "user_input"
    SYSTEM = "system"
    TOOL = "tool"
    LLM = "llm"
    IMPORTED = "imported"


class ArtifactKind(str, enum.Enum):
    FILE = "file"
    LOG = "log"
    REPORT = "report"
    TRACE = "trace"
    IMAGE = "image"
    DATA = "data"


class BillingSource(str, enum.Enum):
    AZURE_MARKETPLACE = "azure_marketplace"
    MANUAL = "manual"


class SubscriptionStatus(str, enum.Enum):
    TRIAL = "trial"
    ACTIVE = "active"
    PAST_DUE = "past_due"
    CANCELLED = "cancelled"
    EXPIRED = "expired"


class Tenant(Base):
    __tablename__ = "tenants"

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    entra_tenant_id: Mapped[str] = mapped_column(
        String(128), nullable=False, unique=True
    )
    display_name: Mapped[str | None] = mapped_column(String(255), nullable=True)
    domain: Mapped[str | None] = mapped_column(String(255), nullable=True)
    plan: Mapped[TenantPlan] = mapped_column(
        _pg_enum(TenantPlan, name="tenant_plan"),
        nullable=False,
        server_default=TenantPlan.FREE.value,
    )
    status: Mapped[TenantStatus] = mapped_column(
        _pg_enum(TenantStatus, name="tenant_status"),
        nullable=False,
        server_default=TenantStatus.ACTIVE.value,
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class User(Base):
    __tablename__ = "users"
    __table_args__ = (
        UniqueConstraint(
            "tenant_id", "entra_user_id", name="uq_users_tenant_entra_user"
        ),
        UniqueConstraint("tenant_id", "id", name="uq_users_tenant_id_id"),
        Index("ix_users_tenant_created_at", "tenant_id", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    entra_user_id: Mapped[str] = mapped_column(String(128), nullable=False)
    email: Mapped[str | None] = mapped_column(String(320), nullable=True)
    full_name: Mapped[str | None] = mapped_column(String(255), nullable=True)
    is_active: Mapped[bool] = mapped_column(
        Boolean, nullable=False, server_default=text("true")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class Membership(Base):
    __tablename__ = "memberships"
    __table_args__ = (
        ForeignKeyConstraint(
            ["tenant_id", "user_id"],
            ["users.tenant_id", "users.id"],
            ondelete="CASCADE",
            name="fk_memberships_tenant_user__users_tenant_id_id",
        ),
        UniqueConstraint("tenant_id", "user_id", name="uq_memberships_tenant_user"),
        Index("ix_memberships_tenant_created_at", "tenant_id", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    user_id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), nullable=False)
    role: Mapped[MembershipRole] = mapped_column(
        _pg_enum(MembershipRole, name="membership_role"),
        nullable=False,
        server_default=MembershipRole.MEMBER.value,
    )
    is_default: Mapped[bool] = mapped_column(
        Boolean, nullable=False, server_default=text("false")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class SandboxSession(Base):
    __tablename__ = "sandbox_sessions"
    __table_args__ = (
        UniqueConstraint("tenant_id", "id", name="uq_sandbox_sessions_tenant_id_id"),
        UniqueConstraint(
            "tenant_id",
            "provider",
            "external_id",
            name="uq_sandbox_sessions_tenant_provider_external",
        ),
        Index("ix_sandbox_sessions_tenant_created_at", "tenant_id", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    provider: Mapped[SandboxProvider] = mapped_column(
        _pg_enum(SandboxProvider, name="sandbox_provider"), nullable=False
    )
    external_id: Mapped[str] = mapped_column(String(255), nullable=False)
    status: Mapped[SandboxSessionStatus] = mapped_column(
        _pg_enum(SandboxSessionStatus, name="sandbox_session_status"),
        nullable=False,
        server_default=SandboxSessionStatus.ACTIVE.value,
    )
    started_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    ended_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    metadata_json: Mapped[dict] = mapped_column(
        "metadata", JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class ModalVolume(Base):
    __tablename__ = "modal_volumes"
    __table_args__ = (
        UniqueConstraint("tenant_id", "id", name="uq_modal_volumes_tenant_id_id"),
        UniqueConstraint(
            "tenant_id", "volume_name", name="uq_modal_volumes_tenant_volume_name"
        ),
        Index("ix_modal_volumes_tenant_created_at", "tenant_id", "created_at"),
        Index("ix_modal_volumes_tenant_last_seen_at", "tenant_id", "last_seen_at"),
        Index(
            "ix_modal_volumes_tenant_provider_env_last_seen",
            "tenant_id",
            "provider",
            "environment",
            "last_seen_at",
        ),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    provider: Mapped[str] = mapped_column(
        String(32), nullable=False, server_default=text("'modal'")
    )
    volume_name: Mapped[str] = mapped_column(String(255), nullable=False)
    external_volume_id: Mapped[str | None] = mapped_column(String(255), nullable=True)
    environment: Mapped[str | None] = mapped_column(String(64), nullable=True)
    region: Mapped[str | None] = mapped_column(String(64), nullable=True)
    metadata_json: Mapped[dict] = mapped_column(
        "metadata", JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    last_seen_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    last_synced_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class Run(Base):
    __tablename__ = "runs"
    __table_args__ = (
        ForeignKeyConstraint(
            ["tenant_id", "created_by_user_id"],
            ["users.tenant_id", "users.id"],
            ondelete="RESTRICT",
            name="fk_runs_tenant_created_by_user__users_tenant_id_id",
        ),
        ForeignKeyConstraint(
            ["tenant_id", "sandbox_session_id"],
            ["sandbox_sessions.tenant_id", "sandbox_sessions.id"],
            ondelete="SET NULL",
            name="fk_runs_tenant_sandbox_session__sandbox_sessions_tenant_id_id",
        ),
        UniqueConstraint(
            "tenant_id", "external_run_id", name="uq_runs_tenant_external_run"
        ),
        UniqueConstraint("tenant_id", "id", name="uq_runs_tenant_id_id"),
        Index("ix_runs_tenant_created_at", "tenant_id", "created_at"),
        Index("ix_runs_tenant_status_created_at", "tenant_id", "status", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    external_run_id: Mapped[str] = mapped_column(String(255), nullable=False)
    created_by_user_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), nullable=False
    )
    status: Mapped[RunStatus] = mapped_column(
        _pg_enum(RunStatus, name="run_status"),
        nullable=False,
        server_default=RunStatus.RUNNING.value,
    )
    model_provider: Mapped[str | None] = mapped_column(String(128), nullable=True)
    model_name: Mapped[str | None] = mapped_column(String(255), nullable=True)
    sandbox_provider: Mapped[SandboxProvider | None] = mapped_column(
        _pg_enum(SandboxProvider, name="sandbox_provider"), nullable=True
    )
    sandbox_session_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    error_json: Mapped[dict | None] = mapped_column(JSONB, nullable=True)
    started_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    completed_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class RunStep(Base):
    __tablename__ = "run_steps"
    __table_args__ = (
        ForeignKeyConstraint(
            ["tenant_id", "run_id"],
            ["runs.tenant_id", "runs.id"],
            ondelete="CASCADE",
            name="fk_run_steps_tenant_run__runs_tenant_id_id",
        ),
        ForeignKeyConstraint(
            ["tenant_id", "modal_volume_id"],
            ["modal_volumes.tenant_id", "modal_volumes.id"],
            ondelete="SET NULL",
            name="fk_run_steps_tenant_modal_volume__modal_volumes_tenant_id_id",
        ),
        UniqueConstraint(
            "tenant_id",
            "run_id",
            "step_index",
            name="uq_run_steps_tenant_run_step_index",
        ),
        UniqueConstraint("tenant_id", "id", name="uq_run_steps_tenant_id_id"),
        Index("ix_run_steps_tenant_created_at", "tenant_id", "created_at"),
        Index("ix_run_steps_run_id", "run_id"),
        Index("ix_run_steps_tenant_run_step", "tenant_id", "run_id", "step_index"),
        Index(
            "ix_run_steps_tenant_type_created_at",
            "tenant_id",
            "step_type",
            "created_at",
        ),
        Index(
            "ix_run_steps_tenant_modal_volume_created_at",
            "tenant_id",
            "modal_volume_id",
            "created_at",
        ),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    run_id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), nullable=False)
    step_index: Mapped[int] = mapped_column(Integer, nullable=False)
    step_type: Mapped[RunStepType] = mapped_column(
        _pg_enum(RunStepType, name="run_step_type"), nullable=False
    )
    input_json: Mapped[dict | None] = mapped_column(JSONB, nullable=True)
    output_json: Mapped[dict | None] = mapped_column(JSONB, nullable=True)
    sandbox_session_external_id: Mapped[str | None] = mapped_column(
        String(255), nullable=True
    )
    modal_volume_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    modal_volume_name: Mapped[str | None] = mapped_column(String(255), nullable=True)
    cost_usd_micros: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    tokens_in: Mapped[int | None] = mapped_column(Integer, nullable=True)
    tokens_out: Mapped[int | None] = mapped_column(Integer, nullable=True)
    latency_ms: Mapped[int | None] = mapped_column(Integer, nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class Artifact(Base):
    __tablename__ = "artifacts"
    __table_args__ = (
        ForeignKeyConstraint(
            ["tenant_id", "run_id"],
            ["runs.tenant_id", "runs.id"],
            ondelete="CASCADE",
            name="fk_artifacts_tenant_run__runs_tenant_id_id",
        ),
        ForeignKeyConstraint(
            ["tenant_id", "step_id"],
            ["run_steps.tenant_id", "run_steps.id"],
            ondelete="SET NULL",
            name="fk_artifacts_tenant_step__run_steps_tenant_id_id",
        ),
        Index("ix_artifacts_tenant_created_at", "tenant_id", "created_at"),
        Index(
            "ix_artifacts_tenant_run_created_at", "tenant_id", "run_id", "created_at"
        ),
        Index("ix_artifacts_tenant_kind_created_at", "tenant_id", "kind", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    run_id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), nullable=False)
    step_id: Mapped[uuid.UUID | None] = mapped_column(UUID(as_uuid=True), nullable=True)
    kind: Mapped[ArtifactKind] = mapped_column(
        _pg_enum(ArtifactKind, name="artifact_kind"), nullable=False
    )
    uri: Mapped[str] = mapped_column(Text, nullable=False)
    mime_type: Mapped[str | None] = mapped_column(String(255), nullable=True)
    size_bytes: Mapped[int | None] = mapped_column(BigInteger, nullable=True)
    checksum: Mapped[str | None] = mapped_column(String(255), nullable=True)
    metadata_json: Mapped[dict] = mapped_column(
        "metadata", JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )


class RLMProgram(Base):
    __tablename__ = "rlm_programs"
    __table_args__ = (
        UniqueConstraint("tenant_id", "id", name="uq_rlm_programs_tenant_id_id"),
        UniqueConstraint("tenant_id", "program_key", name="uq_rlm_programs_tenant_key"),
        Index("ix_rlm_programs_tenant_created_at", "tenant_id", "created_at"),
        Index("ix_rlm_programs_tenant_kind_status", "tenant_id", "kind", "status"),
        Index("ix_rlm_programs_tenant_updated_at", "tenant_id", "updated_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    program_key: Mapped[str] = mapped_column(String(255), nullable=False)
    display_name: Mapped[str | None] = mapped_column(String(255), nullable=True)
    kind: Mapped[str] = mapped_column(
        String(64), nullable=False, server_default=text("'compiled'")
    )
    status: Mapped[str] = mapped_column(
        String(32), nullable=False, server_default=text("'active'")
    )
    dspy_signature: Mapped[str | None] = mapped_column(String(255), nullable=True)
    version_tag: Mapped[str | None] = mapped_column(String(128), nullable=True)
    schema_version: Mapped[int] = mapped_column(
        Integer, nullable=False, server_default=text("1")
    )
    source_run_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    created_by_user_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    program_json: Mapped[dict] = mapped_column(
        JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    metadata_json: Mapped[dict] = mapped_column(
        "metadata", JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class RLMTrace(Base):
    __tablename__ = "rlm_traces"
    __table_args__ = (
        ForeignKeyConstraint(
            ["tenant_id", "run_id"],
            ["runs.tenant_id", "runs.id"],
            ondelete="CASCADE",
            name="fk_rlm_traces_tenant_run__runs_tenant_id_id",
        ),
        ForeignKeyConstraint(
            ["tenant_id", "run_step_id"],
            ["run_steps.tenant_id", "run_steps.id"],
            ondelete="SET NULL",
            name="fk_rlm_traces_tenant_step__run_steps_tenant_id_id",
        ),
        ForeignKeyConstraint(
            ["tenant_id", "program_id"],
            ["rlm_programs.tenant_id", "rlm_programs.id"],
            ondelete="SET NULL",
            name="fk_rlm_traces_tenant_program__rlm_programs_tenant_id_id",
        ),
        UniqueConstraint("tenant_id", "id", name="uq_rlm_traces_tenant_id_id"),
        Index("ix_rlm_traces_tenant_created_at", "tenant_id", "created_at"),
        Index(
            "ix_rlm_traces_tenant_run_created_at", "tenant_id", "run_id", "created_at"
        ),
        Index("ix_rlm_traces_tenant_kind_status", "tenant_id", "trace_kind", "status"),
        Index(
            "ix_rlm_traces_tenant_program_created_at",
            "tenant_id",
            "program_id",
            "created_at",
        ),
        Index(
            "ix_rlm_traces_tenant_step_created_at",
            "tenant_id",
            "run_step_id",
            "created_at",
        ),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    run_id: Mapped[uuid.UUID] = mapped_column(UUID(as_uuid=True), nullable=False)
    run_step_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    program_id: Mapped[uuid.UUID | None] = mapped_column(
        UUID(as_uuid=True), nullable=True
    )
    trace_kind: Mapped[str] = mapped_column(
        String(64), nullable=False, server_default=text("'trajectory'")
    )
    status: Mapped[str] = mapped_column(
        String(32), nullable=False, server_default=text("'captured'")
    )
    source: Mapped[str] = mapped_column(
        String(64), nullable=False, server_default=text("'rlm'")
    )
    summary_text: Mapped[str | None] = mapped_column(Text, nullable=True)
    payload_json: Mapped[dict] = mapped_column(
        JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    tokens_in: Mapped[int | None] = mapped_column(Integer, nullable=True)
    tokens_out: Mapped[int | None] = mapped_column(Integer, nullable=True)
    latency_ms: Mapped[int | None] = mapped_column(Integer, nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class MemoryItem(Base):
    __tablename__ = "memory_items"
    __table_args__ = (
        CheckConstraint(
            "importance >= 0 AND importance <= 100",
            name="ck_memory_items_importance_range",
        ),
        Index("ix_memory_items_tenant_created_at", "tenant_id", "created_at"),
        Index("ix_memory_items_scope", "tenant_id", "scope", "scope_id", "created_at"),
        Index("ix_memory_items_tags", "tags", postgresql_using="gin"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    scope: Mapped[MemoryScope] = mapped_column(
        _pg_enum(MemoryScope, name="memory_scope"), nullable=False
    )
    scope_id: Mapped[str] = mapped_column(String(255), nullable=False)
    kind: Mapped[MemoryKind] = mapped_column(
        _pg_enum(MemoryKind, name="memory_kind"), nullable=False
    )
    content_text: Mapped[str | None] = mapped_column(Text, nullable=True)
    content_json: Mapped[dict | None] = mapped_column(JSONB, nullable=True)
    source: Mapped[MemorySource] = mapped_column(
        _pg_enum(MemorySource, name="memory_source"), nullable=False
    )
    importance: Mapped[int] = mapped_column(
        SmallInteger, nullable=False, server_default=text("0")
    )
    tags: Mapped[list[str]] = mapped_column(
        ARRAY(Text), nullable=False, server_default=text("'{}'::text[]")
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class Job(Base):
    __tablename__ = "jobs"
    __table_args__ = (
        UniqueConstraint(
            "tenant_id", "idempotency_key", name="uq_jobs_tenant_idempotency_key"
        ),
        Index("ix_jobs_status_available_at", "status", "available_at"),
        Index("ix_jobs_tenant_status_available", "tenant_id", "status", "available_at"),
        Index("ix_jobs_tenant_created_at", "tenant_id", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    job_type: Mapped[JobType] = mapped_column(
        _pg_enum(JobType, name="job_type"), nullable=False
    )
    status: Mapped[JobStatus] = mapped_column(
        _pg_enum(JobStatus, name="job_status"),
        nullable=False,
        server_default=JobStatus.QUEUED.value,
    )
    payload: Mapped[dict] = mapped_column(
        JSONB, nullable=False, server_default=text("'{}'::jsonb")
    )
    attempts: Mapped[int] = mapped_column(
        Integer, nullable=False, server_default=text("0")
    )
    max_attempts: Mapped[int] = mapped_column(
        Integer, nullable=False, server_default=text("5")
    )
    available_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    locked_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    locked_by: Mapped[str | None] = mapped_column(String(255), nullable=True)
    idempotency_key: Mapped[str] = mapped_column(String(255), nullable=False)
    last_error: Mapped[dict | None] = mapped_column(JSONB, nullable=True)
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )


class TenantSubscription(Base):
    __tablename__ = "tenant_subscriptions"
    __table_args__ = (
        UniqueConstraint(
            "tenant_id",
            "billing_source",
            "subscription_id",
            name="uq_tenant_subscriptions_source_subscription",
        ),
        Index("ix_tenant_subscriptions_tenant_created_at", "tenant_id", "created_at"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()")
    )
    tenant_id: Mapped[uuid.UUID] = mapped_column(
        UUID(as_uuid=True), ForeignKey("tenants.id", ondelete="CASCADE"), nullable=False
    )
    billing_source: Mapped[BillingSource] = mapped_column(
        _pg_enum(BillingSource, name="billing_source"),
        nullable=False,
        server_default=BillingSource.MANUAL.value,
    )
    subscription_id: Mapped[str] = mapped_column(String(255), nullable=False)
    offer_id: Mapped[str | None] = mapped_column(String(255), nullable=True)
    plan_id: Mapped[str | None] = mapped_column(String(255), nullable=True)
    status: Mapped[SubscriptionStatus] = mapped_column(
        _pg_enum(SubscriptionStatus, name="subscription_status"),
        nullable=False,
        server_default=SubscriptionStatus.ACTIVE.value,
    )
    started_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    ended_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True), nullable=True
    )
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True), nullable=False, server_default=func.now()
    )
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
    )



"""Runtime settings and connectivity diagnostics endpoints."""

from __future__ import annotations

import os
import time
from concurrent.futures import ThreadPoolExecutor, TimeoutError as FutureTimeoutError
from datetime import datetime, timezone
from typing import Any

from fastapi import APIRouter, HTTPException

from fleet_rlm.core.config import get_delegate_lm_from_env, get_planner_lm_from_env
from fleet_rlm.server.runtime_settings import (
    RUNTIME_SETTINGS_ALLOWLIST,
    RUNTIME_SETTINGS_KEYS,
    apply_env_updates,
    get_settings_snapshot,
    normalize_updates,
)
from fleet_rlm.utils.modal import load_modal_config

from ..deps import ServerStateDep
from ..schemas.core import (
    RuntimeConnectivityTestResponse,
    RuntimeActiveModels,
    RuntimeSettingsSnapshot,
    RuntimeSettingsUpdateRequest,
    RuntimeSettingsUpdateResponse,
    RuntimeStatusResponse,
    RuntimeTestCache,
)

router = APIRouter(prefix="/runtime", tags=["runtime"])

_RUNTIME_TEST_TIMEOUT_SECONDS = 20


def _utc_now_iso() -> str:
    return datetime.now(tz=timezone.utc).isoformat()


def _runtime_setting_overrides(
    *, secret_name: str, volume_name: str | None
) -> dict[str, str]:
    return {
        "SECRET_NAME": secret_name,
        "VOLUME_NAME": volume_name or "",
    }


def _resolve_active_model(value: str | None, env_key: str) -> str:
    direct = (value or "").strip()
    if direct:
        return direct
    fallback = (os.environ.get(env_key) or "").strip()
    return fallback


def _sanitize_error(exc: Exception) -> str:
    message = str(exc)
    sensitive_values = [
        os.environ.get("DSPY_LLM_API_KEY"),
        os.environ.get("DSPY_LM_API_KEY"),
        os.environ.get("MODAL_TOKEN_SECRET"),
        os.environ.get("MODAL_TOKEN_ID"),
    ]

    for value in sensitive_values:
        if value and len(value) >= 4:
            message = message.replace(value, "***")

    return message


def _extract_lm_text(response: Any) -> str:
    if isinstance(response, list) and response:
        first = response[0]
        if isinstance(first, dict) and "text" in first:
            return str(first["text"]).strip()
        return str(first).strip()
    return str(response).strip()


def _coerce_output_text(value: Any) -> str:
    if isinstance(value, bytes):
        return value.decode("utf-8", errors="replace").strip()
    return str(value).strip()


def _cache_runtime_test(*, state, result: RuntimeConnectivityTestResponse) -> None:
    state.runtime_test_results[result.kind] = result.model_dump(mode="json")


def _modal_preflight(*, secret_name: str) -> tuple[dict[str, bool], list[str]]:
    modal_cfg = load_modal_config()
    credentials_from_env = bool(
        os.environ.get("MODAL_TOKEN_ID") and os.environ.get("MODAL_TOKEN_SECRET")
    )
    credentials_from_profile = bool(
        modal_cfg.get("token_id") and modal_cfg.get("token_secret")
    )
    credentials_available = credentials_from_env or credentials_from_profile
    secret_name_set = bool(secret_name.strip())

    checks = {
        "credentials_from_env": credentials_from_env,
        "credentials_from_profile": credentials_from_profile,
        "credentials_available": credentials_available,
        "secret_name_set": secret_name_set,
    }

    guidance: list[str] = []
    if not credentials_available:
        guidance.append(
            "Modal credentials missing. Configure MODAL_TOKEN_ID/MODAL_TOKEN_SECRET or run `modal setup`."
        )
    if not secret_name_set:
        guidance.append(
            "Runtime secret name is empty. Set SECRET_NAME in Runtime settings."
        )

    return checks, guidance


def _lm_preflight() -> tuple[dict[str, bool], list[str]]:
    has_model = bool((os.environ.get("DSPY_LM_MODEL") or "").strip())
    has_api_key = bool(
        (
            os.environ.get("DSPY_LLM_API_KEY")
            or os.environ.get("DSPY_LM_API_KEY")
            or ""
        ).strip()
    )
    checks = {
        "model_set": has_model,
        "api_key_set": has_api_key,
    }

    guidance: list[str] = []
    if not has_model:
        guidance.append("DSPY_LM_MODEL is not set.")
    if not has_api_key:
        guidance.append("DSPY_LLM_API_KEY (or DSPY_LM_API_KEY) is not set.")

    return checks, guidance


@router.get("/settings", response_model=RuntimeSettingsSnapshot)
async def get_runtime_settings(state: ServerStateDep) -> RuntimeSettingsSnapshot:
    snapshot = get_settings_snapshot(
        keys=list(RUNTIME_SETTINGS_KEYS),
        extra_values=_runtime_setting_overrides(
            secret_name=state.config.secret_name,
            volume_name=state.config.volume_name,
        ),
    )
    return RuntimeSettingsSnapshot(**snapshot)


@router.patch("/settings", response_model=RuntimeSettingsUpdateResponse)
async def patch_runtime_settings(
    state: ServerStateDep,
    request: RuntimeSettingsUpdateRequest,
) -> RuntimeSettingsUpdateResponse:
    config = state.config
    if config.app_env != "local":
        raise HTTPException(
            status_code=403,
            detail="Runtime settings updates are allowed only when APP_ENV=local.",
        )

    try:
        normalized = normalize_updates(
            request.updates,
            allowlist=RUNTIME_SETTINGS_ALLOWLIST,
        )
    except ValueError as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc

    result = apply_env_updates(updates=normalized)

    if "SECRET_NAME" in normalized:
        resolved_secret_name = normalized["SECRET_NAME"].strip()
        config.secret_name = resolved_secret_name or "LITELLM"

    if "VOLUME_NAME" in normalized:
        resolved_volume_name = normalized["VOLUME_NAME"].strip()
        config.volume_name = resolved_volume_name or None

    if "DSPY_LM_MODEL" in normalized:
        resolved_planner_model = normalized["DSPY_LM_MODEL"].strip()
        config.agent_model = resolved_planner_model or None

    if "DSPY_DELEGATE_LM_MODEL" in normalized:
        resolved_delegate_model = normalized["DSPY_DELEGATE_LM_MODEL"].strip()
        config.agent_delegate_model = resolved_delegate_model or None

    if "DSPY_DELEGATE_LM_SMALL_MODEL" in normalized:
        resolved_delegate_small_model = normalized[
            "DSPY_DELEGATE_LM_SMALL_MODEL"
        ].strip()
        config.agent_delegate_small_model = resolved_delegate_small_model or None

    # Rebuild planner LM in-process to apply updated env values immediately.
    state.planner_lm = get_planner_lm_from_env(model_name=config.agent_model)
    state.delegate_lm = get_delegate_lm_from_env(
        model_name=config.agent_delegate_model,
        default_max_tokens=config.agent_delegate_max_tokens,
    )

    return RuntimeSettingsUpdateResponse(**result)


@router.post("/tests/modal", response_model=RuntimeConnectivityTestResponse)
async def test_modal_connection(
    state: ServerStateDep,
) -> RuntimeConnectivityTestResponse:
    checks, guidance = _modal_preflight(secret_name=state.config.secret_name)
    preflight_ok = checks["credentials_available"] and checks["secret_name_set"]

    checked_at = _utc_now_iso()
    if not preflight_ok:
        result = RuntimeConnectivityTestResponse(
            kind="modal",
            ok=False,
            preflight_ok=False,
            checked_at=checked_at,
            checks=checks,
            guidance=guidance,
            error="Modal preflight checks failed.",
        )
        _cache_runtime_test(state=state, result=result)
        return result

    latency_ms: int | None = None
    output_preview: str | None = None
    error: str | None = None
    ok = False

    sandbox = None
    started = time.perf_counter()
    try:
        import modal

        app = await modal.App.lookup.aio(
            "fleet-rlm-runtime-smoke", create_if_missing=True
        )
        sandbox = await modal.Sandbox.create.aio(app=app, timeout=30)
        proc = await sandbox.exec.aio("python", "-c", "print('ok')", timeout=15)
        await proc.wait.aio()
        output_preview = _coerce_output_text(await proc.stdout.read.aio())
        ok = output_preview == "ok"
        if not ok:
            error = "Modal sandbox returned unexpected output."
    except Exception as exc:  # pragma: no cover - network/provider path
        error = _sanitize_error(exc)
    finally:
        latency_ms = int((time.perf_counter() - started) * 1000)
        if sandbox is not None:
            try:
                await sandbox.terminate.aio()
            except Exception:
                pass

    if not ok and not error:
        error = "Modal connectivity test failed."

    result = RuntimeConnectivityTestResponse(
        kind="modal",
        ok=ok,
        preflight_ok=True,
        checked_at=checked_at,
        checks=checks,
        guidance=guidance,
        latency_ms=latency_ms,
        output_preview=output_preview,
        error=error,
    )
    _cache_runtime_test(state=state, result=result)
    return result


@router.post("/tests/lm", response_model=RuntimeConnectivityTestResponse)
async def test_lm_connection(state: ServerStateDep) -> RuntimeConnectivityTestResponse:
    checks, guidance = _lm_preflight()
    preflight_ok = checks["model_set"] and checks["api_key_set"]

    checked_at = _utc_now_iso()
    if not preflight_ok:
        result = RuntimeConnectivityTestResponse(
            kind="lm",
            ok=False,
            preflight_ok=False,
            checked_at=checked_at,
            checks=checks,
            guidance=guidance,
            error="LM preflight checks failed.",
        )
        _cache_runtime_test(state=state, result=result)
        return result

    latency_ms: int | None = None
    output_preview: str | None = None
    error: str | None = None
    ok = False

    started = time.perf_counter()
    try:
        planner_lm = get_planner_lm_from_env(model_name=state.config.agent_model)
        if planner_lm is None:
            raise RuntimeError(
                "Failed to construct planner LM from environment settings."
            )

        def _invoke() -> str:
            response = planner_lm("Reply with exactly OK")
            return _extract_lm_text(response)

        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(_invoke)
            output_preview = future.result(timeout=_RUNTIME_TEST_TIMEOUT_SECONDS)

        ok = bool(output_preview)
        state.planner_lm = planner_lm
        state.delegate_lm = get_delegate_lm_from_env(
            model_name=state.config.agent_delegate_model,
            default_max_tokens=state.config.agent_delegate_max_tokens,
        )
    except FutureTimeoutError:
        error = (
            f"LM test timed out after {_RUNTIME_TEST_TIMEOUT_SECONDS}s. "
            "Check API connectivity and credentials."
        )
    except Exception as exc:  # pragma: no cover - provider/network path
        error = _sanitize_error(exc)
    finally:
        latency_ms = int((time.perf_counter() - started) * 1000)

    if not ok and not error:
        error = "LM connectivity test failed."

    result = RuntimeConnectivityTestResponse(
        kind="lm",
        ok=ok,
        preflight_ok=True,
        checked_at=checked_at,
        checks=checks,
        guidance=guidance,
        latency_ms=latency_ms,
        output_preview=output_preview,
        error=error,
    )
    _cache_runtime_test(state=state, result=result)
    return result


@router.get("/status", response_model=RuntimeStatusResponse)
async def get_runtime_status(state: ServerStateDep) -> RuntimeStatusResponse:
    llm_checks, llm_guidance = _lm_preflight()
    modal_checks, modal_guidance = _modal_preflight(
        secret_name=state.config.secret_name
    )

    cached_modal = state.runtime_test_results.get("modal")
    cached_lm = state.runtime_test_results.get("lm")

    modal_test = (
        RuntimeConnectivityTestResponse(**cached_modal)
        if isinstance(cached_modal, dict)
        else None
    )
    lm_test = (
        RuntimeConnectivityTestResponse(**cached_lm)
        if isinstance(cached_lm, dict)
        else None
    )

    ready = bool(modal_test and modal_test.ok and lm_test and lm_test.ok)

    guidance: list[str] = []
    guidance.extend(llm_guidance)
    guidance.extend(modal_guidance)
    if modal_test is None or lm_test is None:
        guidance.append(
            "Run Runtime connection tests to validate live provider connectivity."
        )

    return RuntimeStatusResponse(
        app_env=state.config.app_env,
        write_enabled=state.config.app_env == "local",
        ready=ready,
        active_models=RuntimeActiveModels(
            planner=_resolve_active_model(state.config.agent_model, "DSPY_LM_MODEL"),
            delegate=_resolve_active_model(
                state.config.agent_delegate_model, "DSPY_DELEGATE_LM_MODEL"
            ),
            delegate_small=_resolve_active_model(
                state.config.agent_delegate_small_model,
                "DSPY_DELEGATE_LM_SMALL_MODEL",
            ),
        ),
        llm={
            **llm_checks,
            "planner_configured": state.planner_lm is not None,
        },
        modal={
            **modal_checks,
            "secret_name": state.config.secret_name,
            "configured_volume": state.config.volume_name or "",
        },
        tests=RuntimeTestCache(modal=modal_test, lm=lm_test),
        guidance=guidance,
    )



"""Microsoft Entra auth stub provider.

TODO: Implement real JWKS-based token verification for Entra OIDC.
"""

from __future__ import annotations

from fastapi import Request, WebSocket

from .base import AuthError
from .types import NormalizedIdentity


class EntraAuthProvider:
    """Placeholder for future Entra OIDC/JWKS verification."""

    def __init__(
        self,
        *,
        jwks_url: str | None = None,
        issuer: str | None = None,
        audience: str | None = None,
    ) -> None:
        self.jwks_url = jwks_url
        self.issuer = issuer
        self.audience = audience

    async def authenticate_http(self, request: Request) -> NormalizedIdentity:
        raise AuthError(
            "AUTH_MODE=entra is configured, but Entra token verification is not implemented yet.",
            status_code=503,
        )

    async def authenticate_websocket(self, websocket: WebSocket) -> NormalizedIdentity:
        raise AuthError(
            "AUTH_MODE=entra is configured, but Entra token verification is not implemented yet.",
            status_code=503,
        )



"""Terminal UI components for fleet_rlm chat interface.

This module provides UI and rendering components extracted from terminal_chat.py.
All functions are stateless and take required parameters explicitly.
"""

from .commands import (
    _coerce_value,
    _confirm,
    _confirmation_message,
    _normalize_trace_mode,
    _parse_command_payload,
    _print_unknown_command,
    _safe_split,
    _split_slash_command,
    handle_alias_command,
    handle_reset_command,
    handle_slash_command,
    handle_trace_command,
    print_command_palette,
)
from .settings import (
    _resolve_env_path,
    _write_env_updates,
    check_secret,
    check_secret_key,
    run_long_context,
    run_settings,
    settings_llm,
    settings_modal,
)
from .ui import (
    _FleetCompleter,
    _badge,
    _bottom_toolbar,
    _dialog_style,
    _history_path,
    _iter_mention_paths,
    _mask_secret,
    _print_banner,
    _print_result_inline,
    _prompt_choice,
    _prompt_label,
    _prompt_value,
    _render_shell,
    ROLE_STYLES,
)

__all__ = [
    # commands.py
    "_coerce_value",
    "_confirm",
    "_confirmation_message",
    "_normalize_trace_mode",
    "_parse_command_payload",
    "_print_unknown_command",
    "_safe_split",
    "_split_slash_command",
    "handle_alias_command",
    "handle_reset_command",
    "handle_slash_command",
    "handle_trace_command",
    "print_command_palette",
    # settings.py
    "_resolve_env_path",
    "_write_env_updates",
    "check_secret",
    "check_secret_key",
    "run_long_context",
    "run_settings",
    "settings_llm",
    "settings_modal",
    # ui.py
    "_FleetCompleter",
    "_badge",
    "_bottom_toolbar",
    "_dialog_style",
    "_history_path",
    "_iter_mention_paths",
    "_mask_secret",
    "_print_banner",
    "_print_result_inline",
    "_prompt_choice",
    "_prompt_label",
    "_prompt_value",
    "_render_shell",
    "ROLE_STYLES",
]



"""Turn setup and command dispatch helpers for websocket chat."""

from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable
from typing import Any

from fastapi import WebSocket

from fleet_rlm.db import FleetRepository
from fleet_rlm.db.models import RunStatus
from fleet_rlm.db.types import IdentityUpsertResult, RunCreateRequest

from ...execution import ExecutionStepBuilder
from ...utils import parse_model_identity, resolve_sandbox_provider
from .commands import _handle_command
from .helpers import _sanitize_for_log
from .lifecycle import ExecutionLifecycleManager, PersistenceRequiredError

logger = logging.getLogger(__name__)


async def handle_command_with_persist(
    *,
    websocket: WebSocket,
    agent: Any,
    payload: dict[str, Any],
    session_record: dict[str, Any] | None,
    repository: FleetRepository | None,
    identity_rows: IdentityUpsertResult | None,
    persistence_required: bool,
    local_persist: Callable[..., Awaitable[None]],
) -> None:
    """Dispatch command payload and persist session state afterward."""
    await _handle_command(
        websocket,
        agent,
        payload,
        session_record,
        repository=repository,
        identity_rows=identity_rows,
        persistence_required=persistence_required,
    )
    await local_persist(include_volume_save=True)


async def initialize_turn_lifecycle(
    *,
    planner_lm: Any,
    cfg: Any,
    repository: FleetRepository | None,
    identity_rows: IdentityUpsertResult | None,
    persistence_required: bool,
    execution_emitter: Any,
    workspace_id: str,
    user_id: str,
    sess_id: str,
    turn_index: int,
    session_record: dict[str, Any] | None,
) -> tuple[ExecutionLifecycleManager, ExecutionStepBuilder, str, Any]:
    """Create step builder and lifecycle manager for a single message turn."""
    run_id = f"{workspace_id}:{user_id}:{sess_id}:{turn_index}"
    step_builder = ExecutionStepBuilder(run_id=run_id)
    active_run_db_id = None

    if repository is None:
        logger.warning(
            "runtime_persistence_disabled_for_run",
            extra={
                "run_id": run_id,
                "workspace_id": workspace_id,
                "user_id": user_id,
                "session_id": sess_id,
                "code": "persistence_disabled",
            },
        )

    if repository is not None and identity_rows is not None:
        model_provider, model_name = parse_model_identity(
            getattr(planner_lm, "model", None)
        )
        try:
            run_row = await repository.create_run(
                RunCreateRequest(
                    tenant_id=identity_rows.tenant_id,
                    created_by_user_id=identity_rows.user_id,
                    external_run_id=run_id,
                    status=RunStatus.RUNNING,
                    model_provider=model_provider,
                    model_name=model_name,
                    sandbox_provider=resolve_sandbox_provider(cfg.sandbox_provider),
                )
            )
            active_run_db_id = run_row.id
            if session_record is not None:
                session_record["last_run_db_id"] = str(run_row.id)
        except Exception as exc:
            if persistence_required:
                raise PersistenceRequiredError(
                    "run_start_persist_failed",
                    f"Failed to persist run start: {exc}",
                ) from exc
            logger.warning("Failed to persist run start: %s", _sanitize_for_log(exc))

    lifecycle = ExecutionLifecycleManager(
        run_id=run_id,
        workspace_id=workspace_id,
        user_id=user_id,
        session_id=sess_id,
        execution_emitter=execution_emitter,
        step_builder=step_builder,
        repository=repository,
        identity_rows=identity_rows,
        active_run_db_id=active_run_db_id,
        strict_persistence=persistence_required,
        session_record=session_record,
    )
    return lifecycle, step_builder, run_id, active_run_db_id



"""Database package for fleet-rlm Neon persistence."""

from .engine import DatabaseManager, to_async_database_url, to_sync_database_url
from .models import (
    Artifact,
    ArtifactKind,
    Base,
    BillingSource,
    Job,
    JobStatus,
    JobType,
    MemoryItem,
    MemoryKind,
    MemoryScope,
    MemorySource,
    Membership,
    MembershipRole,
    ModalVolume,
    Run,
    RLMProgram,
    RLMTrace,
    RunStatus,
    RunStep,
    RunStepType,
    SandboxProvider,
    SandboxSession,
    SandboxSessionStatus,
    SubscriptionStatus,
    Tenant,
    TenantPlan,
    TenantStatus,
    TenantSubscription,
    User,
)
from .repository import FleetRepository

__all__ = [
    "Base",
    "DatabaseManager",
    "FleetRepository",
    "Tenant",
    "TenantPlan",
    "TenantStatus",
    "User",
    "Membership",
    "MembershipRole",
    "ModalVolume",
    "SandboxSession",
    "SandboxProvider",
    "SandboxSessionStatus",
    "Run",
    "RLMProgram",
    "RLMTrace",
    "RunStatus",
    "RunStep",
    "RunStepType",
    "Artifact",
    "ArtifactKind",
    "MemoryItem",
    "MemoryScope",
    "MemoryKind",
    "MemorySource",
    "Job",
    "JobType",
    "JobStatus",
    "TenantSubscription",
    "BillingSource",
    "SubscriptionStatus",
    "to_async_database_url",
    "to_sync_database_url",
]



"""WebSocket router package.

Canonical WS implementation lives under this package; legacy flat
``ws_*`` modules remain as one-release compatibility shims.
"""

from .api import chat_streaming, execution_stream, router

__all__ = ["router", "chat_streaming", "execution_stream"]



"""URL-backed document source helpers for ReAct document ingestion."""

from __future__ import annotations

import ipaddress
import os
import socket
import tempfile
from pathlib import Path
from typing import Any, Callable
from urllib.parse import urlparse

import httpx

_TEXTUAL_CONTENT_TYPES = (
    "text/",
    "application/json",
    "application/xml",
    "application/javascript",
    "application/x-javascript",
    "application/x-yaml",
    "application/yaml",
)

_URL_SUFFIX_BY_CONTENT_TYPE = {
    "text/html": ".html",
    "application/xhtml+xml": ".html",
    "application/pdf": ".pdf",
    "application/json": ".json",
    "text/plain": ".txt",
    "text/markdown": ".md",
}


def _env_bool(name: str, default: bool) -> bool:
    from fleet_rlm._env_utils import env_bool

    return env_bool(os.getenv(name), default=default)


def _env_int(name: str, default: int) -> int:
    from fleet_rlm._env_utils import env_int

    return env_int(os.getenv(name), default=default)


def is_http_url(value: str) -> bool:
    parsed = urlparse(value)
    return parsed.scheme in {"http", "https"} and bool(parsed.netloc)


def _normalize_content_type(raw_content_type: str | None) -> str:
    if not raw_content_type:
        return ""
    return raw_content_type.split(";", 1)[0].strip().lower()


def _rewrite_gist_page_url(url: str) -> str:
    parsed = urlparse(url)
    if parsed.scheme not in {"http", "https"} or parsed.hostname != "gist.github.com":
        return url

    parts = [part for part in parsed.path.split("/") if part]
    if len(parts) < 2:
        return url

    user, gist_id = parts[0], parts[1]
    if not user or not gist_id:
        return url

    return f"https://gist.githubusercontent.com/{user}/{gist_id}/raw"


def _resolve_host_ips(
    hostname: str, port: int
) -> set[ipaddress.IPv4Address | ipaddress.IPv6Address]:
    infos = socket.getaddrinfo(hostname, port, type=socket.SOCK_STREAM)
    results: set[ipaddress.IPv4Address | ipaddress.IPv6Address] = set()
    for info in infos:
        sockaddr = info[4]
        ip_str = sockaddr[0]
        ip_obj = ipaddress.ip_address(ip_str)
        if isinstance(ip_obj, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            results.add(ip_obj)
    return results


def _assert_public_http_url(url: str, *, allow_private_net: bool) -> None:
    parsed = urlparse(url)
    if parsed.scheme not in {"http", "https"}:
        raise ValueError(
            f"Unsupported URL scheme '{parsed.scheme}'. Only http:// and https:// URLs are supported."
        )
    if not parsed.hostname:
        raise ValueError(f"Invalid URL (missing hostname): {url}")

    hostname = parsed.hostname
    if hostname in {"localhost", "localhost.localdomain"} and not allow_private_net:
        raise ValueError(f"Blocked local/private URL target: {url}")

    if allow_private_net:
        return

    port = parsed.port or (443 if parsed.scheme == "https" else 80)
    try:
        resolved_ips = _resolve_host_ips(hostname, port)
    except socket.gaierror as exc:
        raise ValueError(f"Could not resolve hostname for URL '{url}': {exc}") from exc

    for ip in resolved_ips:
        if not ip.is_global:
            raise ValueError(f"Blocked local/private URL target: {url} ({ip})")


def _infer_download_suffix(url: str, content_type: str) -> str:
    parsed = urlparse(url)
    path_suffix = Path(parsed.path).suffix.lower()
    if path_suffix:
        return path_suffix
    return _URL_SUFFIX_BY_CONTENT_TYPE.get(content_type, ".bin")


def _decode_text_body(body: bytes, response: httpx.Response) -> str:
    encoding = response.encoding or "utf-8"
    try:
        return body.decode(encoding)
    except (LookupError, UnicodeDecodeError):
        return body.decode("utf-8", errors="replace")


def _read_stream_bytes_limited(response: httpx.Response, max_bytes: int) -> bytes:
    chunks: list[bytes] = []
    total = 0
    for chunk in response.iter_bytes():
        total += len(chunk)
        if total > max_bytes:
            raise ValueError(
                f"Fetched URL content exceeds size limit ({max_bytes} bytes)."
            )
        chunks.append(chunk)
    return b"".join(chunks)


def fetch_url_document_content(
    url: str,
    *,
    read_document_content: Callable[[Path], tuple[str, dict[str, Any]]],
) -> tuple[str, dict[str, Any]]:
    """Fetch and extract a document from a public HTTP(S) URL."""
    if not _env_bool("FLEET_URL_DOCUMENT_FETCH_ENABLED", True):
        raise ValueError(
            "URL document loading is disabled. Set FLEET_URL_DOCUMENT_FETCH_ENABLED=true to enable it."
        )

    max_bytes = _env_int("FLEET_URL_DOCUMENT_MAX_BYTES", 10 * 1024 * 1024)
    max_redirects = _env_int("FLEET_URL_DOCUMENT_MAX_REDIRECTS", 5)
    connect_timeout = _env_int("FLEET_URL_DOCUMENT_CONNECT_TIMEOUT_SECONDS", 10)
    read_timeout = _env_int("FLEET_URL_DOCUMENT_TIMEOUT_SECONDS", 20)
    allow_private_net = _env_bool("FLEET_URL_DOCUMENT_ALLOW_PRIVATE_NET", False)
    user_agent = os.getenv(
        "FLEET_URL_DOCUMENT_USER_AGENT", "fleet-rlm/url-document-loader"
    )

    current_url = _rewrite_gist_page_url(url)
    current_redirects = 0
    timeout = httpx.Timeout(
        connect=connect_timeout,
        read=read_timeout,
        write=read_timeout,
        pool=read_timeout,
    )

    with httpx.Client(timeout=timeout, headers={"User-Agent": user_agent}) as client:
        while True:
            _assert_public_http_url(current_url, allow_private_net=allow_private_net)
            try:
                with client.stream(
                    "GET",
                    current_url,
                    follow_redirects=False,
                ) as response:
                    if 300 <= response.status_code < 400:
                        location = response.headers.get("location")
                        if not location:
                            raise ValueError(
                                f"URL redirect from '{current_url}' missing Location header."
                            )
                        if current_redirects >= max_redirects:
                            raise ValueError(
                                f"Too many redirects while fetching URL '{url}'."
                            )
                        current_url = str(response.url.join(location))
                        current_redirects += 1
                        continue

                    if response.status_code >= 400:
                        raise ValueError(
                            f"Failed to fetch URL '{url}' (HTTP {response.status_code})."
                        )

                    final_url = str(response.url)
                    _assert_public_http_url(
                        final_url, allow_private_net=allow_private_net
                    )

                    content_length = response.headers.get("content-length")
                    if content_length is not None:
                        try:
                            declared_length = int(content_length)
                        except ValueError:
                            declared_length = None
                        if declared_length is not None and declared_length > max_bytes:
                            raise ValueError(
                                f"Fetched URL content exceeds size limit ({max_bytes} bytes)."
                            )

                    body = _read_stream_bytes_limited(response, max_bytes=max_bytes)
                    content_type = _normalize_content_type(
                        response.headers.get("content-type")
                    )

                    base_meta: dict[str, Any] = {
                        "source_kind": "url",
                        "source_url": url,
                        "final_url": final_url,
                        "http_status": response.status_code,
                        "content_type": content_type or "application/octet-stream",
                        "fetched_bytes": len(body),
                        "fetch_method": "httpx",
                    }

                    if (
                        content_type.startswith("text/html")
                        or content_type == "application/xhtml+xml"
                    ):
                        suffix = _infer_download_suffix(final_url, content_type)
                        with tempfile.NamedTemporaryFile(
                            suffix=suffix, delete=True
                        ) as tmp:
                            tmp_path = Path(tmp.name)
                            tmp.write(body)
                            tmp.flush()
                            text, extracted_meta = read_document_content(tmp_path)
                        return text, {**base_meta, **extracted_meta}

                    if content_type.startswith(_TEXTUAL_CONTENT_TYPES):
                        return _decode_text_body(body, response), {
                            **base_meta,
                            "source_type": "text",
                            "extraction_method": "httpx_text",
                        }

                    suffix = _infer_download_suffix(final_url, content_type)
                    if suffix in {
                        ".pdf",
                        ".doc",
                        ".docx",
                        ".ppt",
                        ".pptx",
                        ".xls",
                        ".xlsx",
                        ".rtf",
                        ".epub",
                        ".html",
                        ".htm",
                    }:
                        with tempfile.NamedTemporaryFile(
                            suffix=suffix, delete=True
                        ) as tmp:
                            tmp_path = Path(tmp.name)
                            tmp.write(body)
                            tmp.flush()
                            text, extracted_meta = read_document_content(tmp_path)
                        return text, {**base_meta, **extracted_meta}

                    raise ValueError(
                        f"Fetched URL appears to be unsupported/binary content (content-type: {content_type or 'unknown'})."
                    )
            except httpx.TimeoutException as exc:
                raise ValueError(f"Timed out fetching URL '{url}'.") from exc
            except httpx.HTTPError as exc:
                raise ValueError(f"Failed to fetch URL '{url}': {exc}") from exc



"""Timestamp-based chunking strategy for log files."""

from __future__ import annotations

import re


def chunk_by_timestamps(
    text: str,
    pattern: str = r"^\d{4}-\d{2}-\d{2}[T ]",
    flags: int = re.MULTILINE,
) -> list[dict]:
    """Split log-style text by timestamp boundaries.

    Splits text at lines starting with a timestamp pattern. Each chunk
    contains all log entries from one timestamp boundary to the next.

    Args:
        text: The log text to split.
        pattern: Regex pattern matching timestamp line starts.
            Default: ISO-8601 style ``r"^\\d{4}-\\d{2}-\\d{2}[T ]"``.
        flags: Regex flags. Default: ``re.MULTILINE``.

    Returns:
        List of dicts with keys:
            - ``timestamp``: The matched timestamp prefix
            - ``content``: Full log entry/entries for this boundary
            - ``start_pos``: Character offset in original text

    Example:
        >>> logs = "2026-01-01 INFO Start\\n2026-01-02 ERROR Fail"
        >>> chunks = chunk_by_timestamps(logs)
        >>> len(chunks)
        2
    """
    if not text:
        return []

    compiled = re.compile(pattern, flags)
    matches = list(compiled.finditer(text))

    if not matches:
        return [{"timestamp": "", "content": text, "start_pos": 0}]

    chunks: list[dict] = []

    # Content before first timestamp
    if matches[0].start() > 0:
        preamble = text[: matches[0].start()].strip()
        if preamble:
            chunks.append({"timestamp": "", "content": preamble, "start_pos": 0})

    for i, match in enumerate(matches):
        end = matches[i + 1].start() if i + 1 < len(matches) else len(text)
        content = text[match.start() : end].strip()
        timestamp = match.group(0).strip()
        chunks.append(
            {"timestamp": timestamp, "content": content, "start_pos": match.start()}
        )

    return chunks



"""Sub-agent spawning helper for RLM delegate tools.

Provides a unified mechanism for delegate tools to spawn full recursive
sub-agents (``RLMReActChatAgent`` instances) instead of invoking single-shot
``dspy.RLM`` modules directly.  This aligns delegate tools with the true
recursion pattern used by ``rlm_query``.
"""

from __future__ import annotations

import json
import logging
from contextlib import nullcontext
from typing import TYPE_CHECKING, Any

import dspy

if TYPE_CHECKING:
    from .agent import RLMReActChatAgent

logger = logging.getLogger(__name__)


def _claim_delegate_slot_or_error(agent: "RLMReActChatAgent") -> dict[str, Any] | None:
    claim_slot = getattr(agent, "_claim_delegate_slot", None)
    if not callable(claim_slot):
        return None

    claim_result = claim_slot()
    if (
        isinstance(claim_result, tuple)
        and len(claim_result) == 2
        and isinstance(claim_result[0], bool)
    ):
        allowed = bool(claim_result[0])
        limit = int(claim_result[1])
        if not allowed:
            return {
                "status": "error",
                "error": (
                    "Delegate call budget reached for this turn. "
                    f"Maximum delegate calls per turn is {limit}."
                ),
                "delegate_max_calls_per_turn": limit,
            }

    return None


def _build_sub_agent(
    agent: "RLMReActChatAgent", *, delegate_lm: Any | None
) -> "RLMReActChatAgent":
    SubAgentClass = agent.__class__
    return SubAgentClass(
        react_max_iters=getattr(agent, "react_max_iters", 10),
        deep_react_max_iters=getattr(agent, "deep_react_max_iters", 35),
        enable_adaptive_iters=getattr(agent, "enable_adaptive_iters", True),
        rlm_max_iterations=getattr(agent, "rlm_max_iterations", 30),
        rlm_max_llm_calls=getattr(agent, "rlm_max_llm_calls", 50),
        verbose=getattr(agent, "verbose", False),
        history_max_turns=getattr(agent, "history_max_turns", None),
        interpreter=agent.interpreter,
        max_depth=agent._max_depth,
        current_depth=agent._current_depth + 1,
        guardrail_mode=getattr(agent, "guardrail_mode", "off"),
        max_output_chars=getattr(agent, "max_output_chars", 10000),
        min_substantive_chars=getattr(agent, "min_substantive_chars", 20),
        delegate_lm=delegate_lm,
        delegate_max_calls_per_turn=getattr(agent, "delegate_max_calls_per_turn", 8),
        delegate_result_truncation_chars=getattr(
            agent, "delegate_result_truncation_chars", 8000
        ),
    )


def _normalize_delegate_result(
    *,
    agent: "RLMReActChatAgent",
    sub_agent: "RLMReActChatAgent",
    raw_result: dict[str, Any],
    fallback_used: bool,
) -> dict[str, Any]:
    result_copy = dict(raw_result)
    result_copy.setdefault("status", "ok")
    response_text = str(result_copy.get("assistant_response", ""))
    truncation_limit = int(getattr(agent, "delegate_result_truncation_chars", 8000))
    if truncation_limit > 0 and len(response_text) > truncation_limit:
        truncated = response_text[:truncation_limit].rstrip()
        result_copy["assistant_response"] = (
            f"{truncated}\n\n[truncated delegate output]"
        )
        result_copy["delegate_output_truncated"] = True
        if callable(getattr(agent, "_record_delegate_truncation", None)):
            agent._record_delegate_truncation()
    else:
        result_copy["delegate_output_truncated"] = False

    return {
        **result_copy,
        "depth": agent._current_depth + 1,
        "sub_agent_history": sub_agent.history_turns(),
        "delegate_lm_fallback": fallback_used,
    }


async def _run_sub_agent_async(
    *,
    agent: "RLMReActChatAgent",
    sub_agent: "RLMReActChatAgent",
    prompt: str,
    delegate_lm: Any | None,
    parent_lm: Any | None,
    fallback_used: bool,
) -> tuple[dict[str, Any], bool]:
    lm_context = (
        dspy.context(lm=delegate_lm)
        if delegate_lm is not None
        else (dspy.context(lm=parent_lm) if parent_lm is not None else nullcontext())
    )
    try:
        with lm_context:
            result = await sub_agent.achat_turn(prompt)
    except Exception:
        if delegate_lm is not None and parent_lm is not None:
            if callable(getattr(agent, "_record_delegate_fallback", None)):
                agent._record_delegate_fallback()
            fallback_used = True
            with dspy.context(lm=parent_lm):
                result = await sub_agent.achat_turn(prompt)
        else:
            raise

    return result, fallback_used


async def spawn_delegate_sub_agent_async(
    agent: "RLMReActChatAgent",
    *,
    prompt: str,
    document: str | None = None,
    document_alias: str = "active",
) -> dict[str, Any]:
    """Spawn a recursive sub-agent and run a single async chat turn.

    Creates a new ``RLMReActChatAgent`` at ``current_depth + 1``, shares the
    parent interpreter, applies delegate budget/depth guards, and normalizes
    result metadata (fallback + truncation markers).
    """
    if agent._current_depth >= agent._max_depth:
        return {
            "status": "error",
            "error": (
                f"Max recursion depth ({agent._max_depth}) reached. "
                "Cannot spawn delegate sub-agent."
            ),
        }

    budget_error = _claim_delegate_slot_or_error(agent)
    if budget_error is not None:
        return budget_error

    delegate_lm = getattr(agent, "delegate_lm", None)
    parent_lm = getattr(dspy.settings, "lm", None)
    fallback_used = delegate_lm is None
    if fallback_used and callable(getattr(agent, "_record_delegate_fallback", None)):
        agent._record_delegate_fallback()

    sub_agent = _build_sub_agent(agent, delegate_lm=delegate_lm)
    if document is not None:
        sub_agent._set_document(document_alias, document)

    raw_result, fallback_used = await _run_sub_agent_async(
        agent=agent,
        sub_agent=sub_agent,
        prompt=prompt,
        delegate_lm=delegate_lm,
        parent_lm=parent_lm,
        fallback_used=fallback_used,
    )

    return _normalize_delegate_result(
        agent=agent,
        sub_agent=sub_agent,
        raw_result=raw_result,
        fallback_used=fallback_used,
    )


def parse_json_from_response(text: str) -> dict[str, Any] | None:
    """Best-effort extraction of a JSON object from sub-agent prose.

    Tries the full text first, then looks for fenced code blocks.
    Returns ``None`` when no valid JSON object is found.
    """
    text = text.strip()
    try:
        obj = json.loads(text)
        if isinstance(obj, dict):
            return obj
    except (json.JSONDecodeError, ValueError):
        pass  # not valid JSON; fall through to fenced-block / plain-text heuristics

    # Try fenced code block
    for marker in ("```json", "```"):
        start = text.find(marker)
        if start == -1:
            continue
        start += len(marker)
        end = text.find("```", start)
        if end == -1:
            continue
        candidate = text[start:end].strip()
        try:
            obj = json.loads(candidate)
            if isinstance(obj, dict):
                return obj
        except (json.JSONDecodeError, ValueError):
            continue

    return None



"""Execution step construction and parent-link derivation helpers."""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Any, Literal

from .sanitizer import _truncate_text, sanitize_event_payload
from .events import ExecutionStep

ExecutionStepType = Literal["llm", "tool", "repl", "memory", "output"]
ExecutionActorKind = Literal["root_rlm", "sub_agent", "delegate", "unknown"]


def _extract_depth(payload: dict[str, Any]) -> int | None:
    candidates: list[Any] = [
        payload.get("depth"),
        payload.get("delegate_depth"),
        payload.get("sub_agent_depth"),
    ]
    step_data = payload.get("step_data")
    if isinstance(step_data, dict):
        candidates.extend(
            [
                step_data.get("depth"),
                step_data.get("delegate_depth"),
                step_data.get("sub_agent_depth"),
            ]
        )
    for raw in candidates:
        if isinstance(raw, bool):
            continue
        if isinstance(raw, (int, float)):
            return max(0, int(raw))
        if isinstance(raw, str) and raw.isdigit():
            return int(raw)
    return None


def _extract_parent_hint(payload: dict[str, Any]) -> str | None:
    for key in ("parent_step_id", "parent_id"):
        value = payload.get(key)
        if isinstance(value, str) and value.strip():
            return value.strip()
    step_data = payload.get("step_data")
    if isinstance(step_data, dict):
        for key in ("parent_step_id", "parent_id"):
            value = step_data.get(key)
            if isinstance(value, str) and value.strip():
                return value.strip()
    return None


def _iter_actor_sources(payload: dict[str, Any]) -> tuple[dict[str, Any], ...]:
    step_data = payload.get("step_data")
    if isinstance(step_data, dict):
        return (payload, step_data)
    return (payload,)


def _extract_actor_id(payload: dict[str, Any]) -> str | None:
    for source in _iter_actor_sources(payload):
        for key in ("actor_id", "delegate_id", "sub_agent_id", "agent_id"):
            value = source.get(key)
            if isinstance(value, str) and value.strip():
                return value.strip()
    return None


def _extract_actor_kind(
    payload: dict[str, Any],
    *,
    depth: int | None,
) -> ExecutionActorKind:
    def _from_text(value: str) -> ExecutionActorKind | None:
        lowered = value.strip().lower()
        if not lowered:
            return None
        if lowered in {"root", "root_rlm", "root-rlm", "root agent"}:
            return "root_rlm"
        if lowered in {"sub_agent", "sub-agent", "subagent"}:
            return "sub_agent"
        if lowered in {"delegate", "rlm_delegate", "rlm-delegate"}:
            return "delegate"
        return None

    for source in _iter_actor_sources(payload):
        for key in ("actor_kind", "actor", "agent_kind", "agent_role"):
            value = source.get(key)
            if isinstance(value, str):
                mapped = _from_text(value)
                if mapped is not None:
                    return mapped

    execution_profile = str(payload.get("execution_profile", "")).strip().upper()
    if execution_profile == "RLM_DELEGATE":
        return "delegate"

    for source in _iter_actor_sources(payload):
        for key in ("delegate_depth", "delegate_id"):
            value = source.get(key)
            if isinstance(value, (int, float)) or (
                isinstance(value, str) and value.strip()
            ):
                return "delegate"

        for key in ("sub_agent_depth", "sub_agent_id"):
            value = source.get(key)
            if isinstance(value, (int, float)) or (
                isinstance(value, str) and value.strip()
            ):
                return "sub_agent"

    if depth is not None:
        return "sub_agent" if depth > 0 else "root_rlm"

    return "unknown"


def _derive_lane_key(
    actor_kind: ExecutionActorKind,
    actor_id: str | None,
    depth: int | None,
) -> str:
    if actor_id:
        return f"{actor_kind}:{actor_id}"
    if depth is not None:
        return f"{actor_kind}:depth-{depth}"
    return actor_kind


def _extract_tool_name(text: str, payload: dict[str, Any]) -> str | None:
    raw_name = payload.get("tool_name")
    if isinstance(raw_name, str) and raw_name.strip():
        return raw_name.strip()

    status_message = str(payload.get("raw_status", text or "")).strip()
    if status_message.startswith("Calling tool:"):
        return status_message.removeprefix("Calling tool:").strip().split("(", 1)[0]
    if text.startswith("tool call:"):
        return text.removeprefix("tool call:").strip().split("(", 1)[0]
    return None


def _tool_step_type(tool_name: str | None) -> ExecutionStepType:
    if not tool_name:
        return "tool"
    lowered = tool_name.lower()
    if lowered.startswith("memory_") or lowered.startswith("core_memory_"):
        return "memory"
    return "tool"


@dataclass(slots=True)
class ExecutionStepBuilder:
    """Deterministic step-id and parent-link builder for one execution run."""

    run_id: str
    root_id: str = field(init=False)
    _counter: int = 0
    _last_tool_step_id: str | None = None
    _depth_parents: dict[int, str] = field(default_factory=dict)
    _repl_parent_by_hash: dict[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.root_id = f"{self.run_id}:root"
        self._depth_parents[0] = self.root_id

    def _next_id(self) -> str:
        self._counter += 1
        return f"{self.run_id}:s{self._counter}"

    def _resolve_parent(self, payload: dict[str, Any]) -> str:
        parent_hint = _extract_parent_hint(payload)
        if parent_hint:
            return parent_hint

        depth = _extract_depth(payload)
        if depth is None:
            return self.root_id
        if depth <= 0:
            return self.root_id
        return self._depth_parents.get(depth - 1, self.root_id)

    def _remember_depth_parent(self, payload: dict[str, Any], step_id: str) -> None:
        depth = _extract_depth(payload)
        if depth is not None:
            self._depth_parents[depth] = step_id

    def _build_step(
        self,
        *,
        step_type: ExecutionStepType,
        label: str,
        input_payload: Any,
        output_payload: Any,
        timestamp: float,
        parent_id: str | None = None,
        raw_payload: dict[str, Any] | None = None,
    ) -> ExecutionStep:
        resolved_parent_id = parent_id or self.root_id
        depth = _extract_depth(raw_payload or {})
        actor_kind = _extract_actor_kind(raw_payload or {}, depth=depth)
        if actor_kind == "unknown" and resolved_parent_id == self.root_id:
            actor_kind = "root_rlm"
            if depth is None:
                depth = 0
        actor_id = _extract_actor_id(raw_payload or {})
        step = ExecutionStep(
            id=self._next_id(),
            parent_id=resolved_parent_id,
            type=step_type,
            label=_truncate_text(label),
            depth=depth,
            actor_kind=actor_kind,
            actor_id=actor_id,
            lane_key=_derive_lane_key(actor_kind, actor_id, depth),
            input=sanitize_event_payload(input_payload),
            output=sanitize_event_payload(output_payload),
            timestamp=timestamp,
        )
        if raw_payload is not None:
            self._remember_depth_parent(raw_payload, step.id)
        return step

    def from_stream_event(
        self,
        *,
        kind: str,
        text: str,
        payload: dict[str, Any] | None,
        timestamp: float,
    ) -> ExecutionStep | None:
        payload_obj = payload if isinstance(payload, dict) else {}
        stripped_text = (text or "").strip()

        if kind == "assistant_token":
            return None

        if kind == "status":
            if not stripped_text:
                return None
            if (
                stripped_text.startswith("Calling tool:")
                or stripped_text == "Tool finished."
            ):
                return None
            return self._build_step(
                step_type="llm",
                label=stripped_text,
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "reasoning_step":
            return self._build_step(
                step_type="llm",
                label=stripped_text or "reasoning",
                input_payload=payload_obj or {"event_kind": kind},
                output_payload={"text": stripped_text},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "tool_call":
            tool_name = _extract_tool_name(stripped_text, payload_obj)
            step = self._build_step(
                step_type=_tool_step_type(tool_name),
                label=tool_name or stripped_text or "tool_call",
                input_payload=payload_obj,
                output_payload=None,
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )
            self._last_tool_step_id = step.id
            return step

        if kind == "tool_result":
            tool_name = _extract_tool_name(stripped_text, payload_obj)
            parent_id = self._last_tool_step_id or self._resolve_parent(payload_obj)
            return self._build_step(
                step_type=_tool_step_type(tool_name),
                label=(tool_name or stripped_text or "tool_result"),
                input_payload={"event_kind": kind, "tool_name": tool_name},
                output_payload=payload_obj,
                timestamp=timestamp,
                parent_id=parent_id,
                raw_payload=payload_obj,
            )

        if kind == "trajectory_step":
            step_data = payload_obj.get("step_data")
            step_dict = step_data if isinstance(step_data, dict) else {}
            tool_name = _extract_tool_name(stripped_text, step_dict)
            label = (
                tool_name
                or str(step_dict.get("thought", "")).strip()
                or str(step_dict.get("action", "")).strip()
                or stripped_text
                or "trajectory_step"
            )
            return self._build_step(
                step_type=_tool_step_type(tool_name) if tool_name else "llm",
                label=label,
                input_payload=step_dict.get("input", step_dict),
                output_payload=step_dict.get("output", step_dict.get("observation")),
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "final":
            return self._build_step(
                step_type="output",
                label="assistant_output",
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text, "payload": payload_obj},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind in {"cancelled", "error"}:
            return self._build_step(
                step_type="output",
                label=kind,
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text, "payload": payload_obj},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "plan_update":
            return self._build_step(
                step_type="llm",
                label="plan_update",
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text, "payload": payload_obj},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "rlm_executing":
            return self._build_step(
                step_type="repl",
                label="rlm_executing",
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text, "payload": payload_obj},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        if kind == "memory_update":
            return self._build_step(
                step_type="memory",
                label="memory_update",
                input_payload={"event_kind": kind},
                output_payload={"text": stripped_text, "payload": payload_obj},
                timestamp=timestamp,
                parent_id=self._resolve_parent(payload_obj),
                raw_payload=payload_obj,
            )

        return None

    def from_interpreter_hook(self, payload: dict[str, Any]) -> ExecutionStep | None:
        if not isinstance(payload, dict):
            return None
        phase = str(payload.get("phase", "")).strip().lower()
        timestamp_raw = payload.get("timestamp", time.time())
        try:
            timestamp = float(timestamp_raw)
        except (TypeError, ValueError):
            timestamp = time.time()

        code_hash = str(payload.get("code_hash", "")).strip()
        if phase == "start":
            step = self._build_step(
                step_type="repl",
                label="repl_execute",
                input_payload=payload,
                output_payload=None,
                timestamp=timestamp,
                parent_id=self.root_id,
                raw_payload=payload,
            )
            if code_hash:
                self._repl_parent_by_hash[code_hash] = step.id
            return step

        if phase == "complete":
            parent_id = self._repl_parent_by_hash.pop(code_hash, None) or self.root_id
            return self._build_step(
                step_type="repl",
                label="repl_result",
                input_payload={"event_kind": "repl_complete", "code_hash": code_hash},
                output_payload=payload,
                timestamp=timestamp,
                parent_id=parent_id,
                raw_payload=payload,
            )

        return None
